import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import tensorflow as tf
import numpy as np
from tqdm import tqdm
import re
import collections
import json
import os
from tensorflow.contrib import seq2seq
from sklearn.metrics import pairwise_distances_argmin_min
from sklearn.cluster import KMeans
from .paths import PATH_SUMMARIZE, S3_PATH_SUMMARIZE
from .text_functions import split_by_dot, summary_textcleaning
from .stemmer import sastrawi_stemmer
from .utils import download_file, load_graph
import uuid


def sequence(s, w2v_model, maxlen = 50, vocabulary_size = 50000):
    words = s.split()
    np_array = np.zeros((maxlen), dtype = np.int32)
    current_no = 0
    for no, word in enumerate(words[: maxlen - 2]):
        id_to_append = 1
        if word in w2v_model:
            word_id = w2v_model[word]
            if word_id < vocabulary_size:
                id_to_append = word_id
        np_array[no] = id_to_append
        current_no = no
    np_array[current_no + 1] = 3
    return np_array


def batch_sequence(sentences, dictionary, maxlen = 50):
    np_array = np.zeros((len(sentences), maxlen), dtype = np.int32)
    for no_sentence, sentence in enumerate(sentences):
        current_no = 0
        for no, word in enumerate(sentence.split()[: maxlen - 2]):
            val = dictionary[word] if word in dictionary else 1
            np_array[no_sentence, no] = val
            current_no = no
        np_array[no_sentence, current_no + 1] = 3
    return np_array


class DEEP_SUMMARIZER:
    def __init__(self, sess, x, logits, w2v):
        self._sess = sess
        self._X = x
        self._logits = logits
        self.w2v = w2v

    def summarize(self, corpus, top_k = 3):
        """
        Summarize list of strings / corpus

        Parameters
        ----------
        corpus: str, list

        top_k: int, (default=3)
            number of summarized strings

        Returns
        -------
        string: summarized string
        """
        assert (
            isinstance(corpus, list) and isinstance(corpus[0], str)
        ) or isinstance(
            corpus, str
        ), 'input must be list of strings or a string'
        if isinstance(corpus, str):
            corpus = split_by_dot(corpus)
        else:
            corpus = ' '.join(corpus)
            corpus = re.findall('(?=\S)[^.\n]+(?<=\S)', corpus)

        corpus = [summary_textcleaning(i) for i in corpus]

        sequences = np.array(
            [sequence(sentence, self.w2v) for sentence in corpus]
        )
        encoded = self._sess.run(
            self._logits, feed_dict = {self._X: np.array(sequences)}
        )
        kmeans = KMeans(n_clusters = top_k, random_state = 0)
        kmeans = kmeans.fit(encoded)
        avg = []
        for j in range(top_k):
            idx = np.where(kmeans.labels_ == j)[0]
            avg.append(np.mean(idx))
        closest, _ = pairwise_distances_argmin_min(
            kmeans.cluster_centers_, encoded
        )
        ordering = sorted(range(top_k), key = lambda k: avg[k])
        return '. '.join([corpus[closest[idx]] for idx in ordering])


class Model:
    def __init__(
        self,
        vocabulary_size,
        maxlen = 50,
        output_size = 512,
        learning_rate = 1e-3,
        embedding_size = 256,
        batch_size = 16,
        max_grad_norm = 10,
    ):
        special_embeddings = tf.Variable(
            tf.random_uniform([4, embedding_size], -np.sqrt(3), np.sqrt(3))
        )
        word_embeddings = tf.Variable(
            tf.random_uniform(
                [vocabulary_size, embedding_size], -np.sqrt(3), np.sqrt(3)
            )
        )
        self.output_size = output_size
        self.maxlen = maxlen
        self.embeddings = tf.concat([special_embeddings, word_embeddings], 0)
        self.output_layer = tf.layers.Dense(vocabulary_size)
        self.output_layer.build(output_size)

        self.BEFORE = tf.placeholder(tf.int32, [None, maxlen])
        self.INPUT = tf.placeholder(tf.int32, [None, maxlen])
        self.AFTER = tf.placeholder(tf.int32, [None, maxlen])
        self.batch_size = tf.shape(self.INPUT)[0]

        self.get_thought = self.thought(self.INPUT)
        self.fw_logits = self.decoder(self.get_thought, self.AFTER)
        self.bw_logits = self.decoder(self.get_thought, self.BEFORE)
        self.loss = self.calculate_loss(
            self.fw_logits, self.AFTER
        ) + self.calculate_loss(self.bw_logits, self.BEFORE)
        tvars = tf.trainable_variables()
        grads, _ = tf.clip_by_global_norm(
            tf.gradients(self.loss, tvars), max_grad_norm
        )
        self.optimizer = tf.train.AdamOptimizer(learning_rate).apply_gradients(
            zip(grads, tvars)
        )

    def get_embedding(self, inputs):
        return tf.nn.embedding_lookup(self.embeddings, inputs)

    def thought(self, inputs):
        encoder_in = self.get_embedding(inputs)
        fw_cell = tf.nn.rnn_cell.GRUCell(self.output_size)
        bw_cell = tf.nn.rnn_cell.GRUCell(self.output_size)
        sequence_length = tf.reduce_sum(tf.sign(inputs), axis = 1)
        with tf.variable_scope(
            'thought_scope_' + str(uuid.uuid4()), reuse = False
        ):
            rnn_output = tf.nn.bidirectional_dynamic_rnn(
                fw_cell,
                bw_cell,
                encoder_in,
                sequence_length = sequence_length,
                dtype = tf.float32,
            )[1]
            return sum(rnn_output)

    def decoder(self, thought, labels):
        main = tf.strided_slice(labels, [0, 0], [self.batch_size, -1], [1, 1])
        shifted_labels = tf.concat([tf.fill([self.batch_size, 1], 2), main], 1)
        decoder_in = self.get_embedding(shifted_labels)
        cell = tf.nn.rnn_cell.GRUCell(self.output_size)
        max_seq_lengths = tf.fill([self.batch_size], self.maxlen)
        helper = seq2seq.TrainingHelper(
            decoder_in, max_seq_lengths, time_major = False
        )
        decoder = seq2seq.BasicDecoder(cell, helper, thought)
        decoder_out = seq2seq.dynamic_decode(decoder)[0].rnn_output
        return decoder_out

    def calculate_loss(self, outputs, labels):
        mask = tf.cast(tf.sign(labels), tf.float32)
        logits = self.output_layer(outputs)
        return seq2seq.sequence_loss(logits, labels, mask)


def counter_words(sentences):
    word_counter = collections.Counter()
    word_list = []
    num_lines, num_words = (0, 0)
    for i in sentences:
        words = re.findall('[\\w\']+|[;:\-\(\)&.,!?"]', i)
        word_counter.update(words)
        word_list.extend(words)
        num_lines += 1
        num_words += len(words)
    return word_counter, word_list, num_lines, num_words


def build_dict(word_counter, vocab_size = 50000):
    count = [['PAD', 0], ['UNK', 1], ['START', 2], ['END', 3]]
    count.extend(word_counter.most_common(vocab_size))
    dictionary = dict()
    for word, _ in count:
        dictionary[word] = len(dictionary)
    return dictionary, {word: idx for idx, word in dictionary.items()}


def load_model():
    if not os.path.isfile(PATH_SUMMARIZE['model']):
        print('downloading SUMMARIZE skip-thought frozen model')
        download_file(S3_PATH_SUMMARIZE['model'], PATH_SUMMARIZE['model'])
    if not os.path.isfile(PATH_SUMMARIZE['setting']):
        print('downloading SUMMARIZE skip-thought dictionary')
        download_file(S3_PATH_SUMMARIZE['setting'], PATH_SUMMARIZE['setting'])
    g = load_graph(PATH_SUMMARIZE['model'])
    x = g.get_tensor_by_name('import/Placeholder_1:0')
    logits = g.get_tensor_by_name('import/add_1:0')
    sess = tf.InteractiveSession(graph = g)
    with open(PATH_SUMMARIZE['setting']) as fopen:
        dictionary = json.load(fopen)
    return DEEP_SUMMARIZER(sess, x, logits, dictionary)


def train_model(
    train_X,
    train_Y,
    epoch = 10,
    batch_size = 16,
    embedding_size = 256,
    output_size = 300,
    maxlen = 100,
):
    word_counter, _, _, _ = counter_words(train_X)
    dictionary, _ = build_dict(word_counter)
    sess = tf.InteractiveSession()
    model = Model(
        len(dictionary),
        embedding_size = embedding_size,
        output_size = output_size,
        batch_size = batch_size,
        maxlen = maxlen,
    )
    sess.run(tf.global_variables_initializer())
    for e in range(epoch):
        pbar = tqdm(range(0, len(train_X), batch_size), desc = 'minibatch loop')
        for i in pbar:
            batch_x = batch_sequence(
                train_X[i : min(i + batch_size, len(train_X))],
                dictionary,
                maxlen = maxlen,
            )
            batch_y = batch_sequence(
                train_Y[i : min(i + batch_size, len(train_X))],
                dictionary,
                maxlen = maxlen,
            )
            loss, _ = sess.run(
                [model.loss, model.optimizer],
                feed_dict = {
                    model.BEFORE: batch_y,
                    model.INPUT: batch_x,
                    model.AFTER: batch_y,
                },
            )
            pbar.set_postfix(cost = loss)
    return sess, model, dictionary



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import re
import os
import numpy as np
import itertools
import collections
import random
from unidecode import unidecode
from .utils import download_file
from .tatabahasa import stopword_tatabahasa
from . import home

STOPWORDS = None
stopwords_location = home + '/stop-word-kerulnet'
if not os.path.isfile(stopwords_location):
    print('downloading stopwords')
    download_file('stop-word-kerulnet', stopwords_location)
with open(stopwords_location, 'r') as fopen:
    STOPWORDS = list(filter(None, fopen.read().split('\n')))

STOPWORDS = list(set(STOPWORDS + stopword_tatabahasa))
VOWELS = 'aeiou'
PHONES = ['sh', 'ch', 'ph', 'sz', 'cz', 'sch', 'rz', 'dz']


def _isWord(word):
    if word:
        consecutiveVowels = 0
        consecutiveConsonents = 0
        for idx, letter in enumerate(word.lower()):
            vowel = True if letter in VOWELS else False
            if idx:
                prev = word[idx - 1]
                prevVowel = True if prev in VOWELS else False
                if not vowel and letter == 'y' and not prevVowel:
                    vowel = True
                if prevVowel != vowel:
                    consecutiveVowels = 0
                    consecutiveConsonents = 0
            if vowel:
                consecutiveVowels += 1
            else:
                consecutiveConsonents += 1
            if consecutiveVowels >= 3 or consecutiveConsonents > 3:
                return False
            if consecutiveConsonents == 3:
                subStr = word[idx - 2 : idx + 1]
                if any(phone in subStr for phone in PHONES):
                    consecutiveConsonents -= 1
                    continue
                return False
    return True


_list_laughing = [
    'huhu',
    'haha',
    'gaga',
    'hihi',
    'wkawka',
    'wkwk',
    'kiki',
    'keke',
    'huehue',
]


def malaya_textcleaning(string):
    """
    use by normalizer, spell
    remove links, hashtags, alias
    only accept A-Z, a-z
    remove any laugh
    remove any repeated char more than 2 times
    remove most of nonsense words
    """
    string = re.sub(
        'http\S+|www.\S+',
        '',
        ' '.join(
            [
                word
                for word in string.split()
                if word.find('#') < 0 and word.find('@') < 0
            ]
        ),
    )
    string = unidecode(string).replace('.', '. ')
    string = string.replace(',', ', ')
    string = re.sub('[^\'"A-Za-z\- ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string.lower()).strip()
    string = [word for word in string.lower().split() if _isWord(word)]
    string = [
        word
        for word in string
        if not any([laugh in word for laugh in _list_laughing])
        and word[: len(word) // 2] != word[len(word) // 2 :]
    ]
    string = ' '.join(string)
    string = (
        ''.join(''.join(s)[:2] for _, s in itertools.groupby(string))
    ).split()
    return ' '.join([word for word in string if word not in STOPWORDS])


def normalizer_textcleaning(string):
    """
    use by normalizer, spell
    remove links, hashtags, alias
    only accept A-Z, a-z
    remove any laugh
    remove any repeated char more than 2 times
    """
    string = re.sub(
        'http\S+|www.\S+',
        '',
        ' '.join(
            [
                word
                for word in string.split()
                if word.find('#') < 0 and word.find('@') < 0
            ]
        ),
    )
    string = re.sub('[^A-Za-z ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string).strip()
    string = [
        word.title() if word.isupper() else word
        for word in string.split()
        if len(word)
    ]
    string = [
        word
        for word in string
        if not any([laugh in word for laugh in _list_laughing])
    ]
    string = ' '.join(string)
    return ''.join(''.join(s)[:2] for _, s in itertools.groupby(string))


def simple_textcleaning(string):
    """
    use by topic modelling
    only accept A-Z, a-z
    """
    string = unidecode(string)
    string = re.sub('[^A-Za-z ]+', ' ', string)
    return re.sub(r'[ ]+', ' ', string.lower()).strip()


def entities_textcleaning(string):
    """
    use by xgb entities, multinomial entities,
    xgb pos, xgb entities, char model, word model, concat model
    """
    string = re.sub('[^A-Za-z0-9\-\/ ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string).strip()
    return [
        word.title() if word.isupper() else word
        for word in string.split()
        if len(word)
    ]


def summary_textcleaning(string):
    string = re.sub('[^A-Za-z0-9\-\/\'"\.\, ]+', ' ', string)
    return re.sub(r'[ ]+', ' ', string.lower()).strip()


def get_hashtags(string):
    return [hash.lower() for hash in re.findall('#(\w+)', string)]


def split_by_dot(string):
    string = re.sub(
        r'(?<!\d)\.(?!\d)',
        'SPLITTT',
        string.replace('\n', '').replace('/', ' '),
    )
    string = string.split('SPLITTT')
    return [re.sub(r'[ ]+', ' ', sentence).strip() for sentence in string]


def cluster_pos(result):
    output = {
        'ADJ': [],
        'ADP': [],
        'ADV': [],
        'ADX': [],
        'CCONJ': [],
        'DET': [],
        'NOUN': [],
        'NUM': [],
        'PART': [],
        'PRON': [],
        'PROPN': [],
        'SCONJ': [],
        'SYM': [],
        'VERB': [],
        'X': [],
    }
    last_label, words = None, []
    for word, label in result:
        if last_label != label and last_label:
            joined = ' '.join(words)
            if joined not in output[last_label]:
                output[last_label].append(joined)
            words = []
            last_label = label
            words.append(word)

        else:
            if not last_label:
                last_label = label
            words.append(word)
    return output


def cluster_entities(result):
    output = {
        'OTHER': [],
        'law': [],
        'location': [],
        'organization': [],
        'person': [],
        'quantity': [],
        'time': [],
        'event': [],
    }
    last_label, words = None, []
    for word, label in result:
        if last_label != label and last_label:
            joined = ' '.join(words)
            if joined not in output[last_label]:
                output[last_label].append(joined)
            words = []
            last_label = label
            words.append(word)

        else:
            if not last_label:
                last_label = label
            words.append(word)
    return output


def pos_entities_textcleaning(string):
    """
    use by text entities and pos
    remove links, hashtags, alias
    """
    string = re.sub(
        'http\S+|www.\S+',
        '',
        ' '.join(
            [i for i in string.split() if i.find('#') < 0 and i.find('@') < 0]
        ),
    )
    string = unidecode(string).replace('.', '. ').replace(',', ', ')
    string = re.sub('[^A-Za-z\- ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string).strip()
    return ' '.join(
        [
            word.title() if word.isupper() else word
            for word in string.split()
            if len(word)
        ]
    )


def classification_textcleaning(string, no_stopwords = False, lowering = True):
    """
    use by our text classifiers, stemmer, summarization, topic-modelling
    remove links, hashtags, alias
    """
    string = re.sub(
        'http\S+|www.\S+',
        '',
        ' '.join(
            [i for i in string.split() if i.find('#') < 0 and i.find('@') < 0]
        ),
    )
    string = unidecode(string).replace('.', '. ').replace(',', ', ')
    string = re.sub('[^A-Za-z ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string).strip()
    if no_stopwords:
        string = ' '.join(
            [
                i
                for i in re.findall('[\\w\']+|[;:\-\(\)&.,!?"]', string)
                if len(i)
            ]
        )
        if lowering:
            return string.lower()
        else:
            return ' '.join(
                [
                    word.title() if word.isupper() else word
                    for word in string.split()
                    if len(word)
                ]
            )
    else:
        string = ' '.join(
            [
                i
                for i in re.findall('[\\w\']+|[;:\-\(\)&.,!?"]', string)
                if len(i) and i not in STOPWORDS
            ]
        )
        if lowering:
            return string.lower()
        else:
            return ' '.join(
                [
                    word.title() if word.isupper() else word
                    for word in string.split()
                    if len(word)
                ]
            )


def process_word_pos_entities(word):
    word = word.lower()
    word = re.sub('[^A-Za-z0-9\- ]+', '', word)
    if word.isdigit():
        word = 'NUM'
    return word


def separate_dataset(trainset):
    datastring = []
    datatarget = []
    for i in range(len(trainset.data)):
        data_ = trainset.data[i].split('\n')
        data_ = list(filter(None, data_))
        datastring += data_
        for n in range(len(data_)):
            datatarget.append(trainset.target[i])
    return datastring, datatarget


def print_topics_modelling(
    topics, feature_names, sorting, topics_per_chunk = 6, n_words = 20
):
    for i in range(0, len(topics), topics_per_chunk):
        these_topics = topics[i : i + topics_per_chunk]
        len_this_chunk = len(these_topics)
        print(('topic {:<8}' * len_this_chunk).format(*these_topics))
        print(('-------- {0:<5}' * len_this_chunk).format(''))
        for i in range(n_words):
            print(
                ('{:<14}' * len_this_chunk).format(
                    *feature_names[sorting[these_topics, i]]
                )
            )
        print('\n')


def cluster_words(list_words):
    dict_words = {}
    for word in list_words:
        found = False
        for key in dict_words.keys():
            if word in key or any(
                [word in inside for inside in dict_words[key]]
            ):
                dict_words[key].append(word)
                found = True
            if key in word:
                dict_words[key].append(word)
        if not found:
            dict_words[word] = [word]
    results = []
    for key, words in dict_words.items():
        results.append(max(list(set([key] + words)), key = len))
    return list(set(results))


def str_idx(corpus, dic, maxlen, UNK = 0):
    X = np.zeros((len(corpus), maxlen))
    for i in range(len(corpus)):
        for no, k in enumerate(corpus[i].split()[:maxlen][::-1]):
            X[i, -1 - no] = dic.get(k, UNK)
    return X


def stemmer_str_idx(corpus, dic, UNK = 3):
    X = []
    for i in corpus:
        ints = []
        for k in i:
            ints.append(dic.get(k, UNK))
        X.append(ints)
    return X


def fasttext_str_idx(texts, dictionary):
    idx_trainset = []
    for text in texts:
        idx = []
        for t in text.split():
            if t in dictionary:
                idx.append(dictionary[t])
        idx_trainset.append(idx)
    return idx_trainset


def pad_sentence_batch(sentence_batch, pad_int):
    padded_seqs = []
    seq_lens = []
    max_sentence_len = max([len(sentence) for sentence in sentence_batch])
    for sentence in sentence_batch:
        padded_seqs.append(
            sentence + [pad_int] * (max_sentence_len - len(sentence))
        )
        seq_lens.append(len(sentence))
    return padded_seqs, seq_lens


def pad_sequence(
    sequence,
    n,
    pad_left = False,
    pad_right = False,
    left_pad_symbol = None,
    right_pad_symbol = None,
):
    sequence = iter(sequence)
    if pad_left:
        sequence = itertools.chain((left_pad_symbol,) * (n - 1), sequence)
    if pad_right:
        sequence = itertools.chain(sequence, (right_pad_symbol,) * (n - 1))
    return sequence


def ngrams(
    sequence,
    n,
    pad_left = False,
    pad_right = False,
    left_pad_symbol = None,
    right_pad_symbol = None,
):
    sequence = pad_sequence(
        sequence, n, pad_left, pad_right, left_pad_symbol, right_pad_symbol
    )

    history = []
    while n > 1:
        try:
            next_item = next(sequence)
        except StopIteration:
            return
        history.append(next_item)
        n -= 1
    for item in sequence:
        history.append(item)
        yield tuple(history)
        del history[0]


def add_ngram(sequences, token_indice, ngram = (2, 3)):
    new_sequences = []
    for input_list in sequences:
        new_list = input_list[:]
        for ngram_value in range(ngram[0], ngram[1]):
            for i in range(len(new_list) - ngram_value + 1):
                ngram = tuple(new_list[i : i + ngram_value])
                if ngram in token_indice:
                    new_list.append(token_indice[ngram])
        new_sequences.append(new_list)
    return new_sequences


def generate_ngram(
    result_pos,
    result_entities,
    ngram = (1, 3),
    accept_pos = ['NOUN', 'PROPN', 'VERB'],
    accept_entities = ['law', 'location', 'organization', 'person', 'time'],
):
    words = []
    sentences = []
    for no in range(len(result_pos)):
        if (
            result_pos[no][1] in accept_pos
            or result_entities[no][1] in accept_entities
        ):
            words.append(result_pos[no][0])
    for gram in range(ngram[0], ngram[1] + 1, 1):
        gram_words = list(ngrams(words, gram))
        for sentence in gram_words:
            sentences.append(' '.join(sentence))
    return list(set(sentences))


def sentence_ngram(sentence, ngram = (1, 3)):
    words = sentence.split()
    sentences = []
    for gram in range(ngram[0], ngram[1] + 1, 1):
        gram_words = list(ngrams(words, gram))
        for sentence in gram_words:
            sentences.append(' '.join(sentence))
    return list(set(sentences))


def char_str_idx(corpus, dic, UNK = 0):
    maxlen = max([len(i) for i in corpus])
    X = np.zeros((len(corpus), maxlen))
    for i in range(len(corpus)):
        for no, k in enumerate(corpus[i][:maxlen][::-1]):
            X[i, -1 - no] = dic.get(k, UNK)
    return X


def generate_char_seq(batch, idx2word, char2idx):
    x = [[len(idx2word[i]) for i in k] for k in batch]
    maxlen = max([j for i in x for j in i])
    temp = np.zeros((batch.shape[0], batch.shape[1], maxlen), dtype = np.int32)
    for i in range(batch.shape[0]):
        for k in range(batch.shape[1]):
            for no, c in enumerate(idx2word[batch[i, k]].lower()):
                temp[i, k, -1 - no] = char2idx[c]
    return temp


def build_dataset(words, n_words, included_prefix = True):
    count = (
        [['GO', 0], ['PAD', 1], ['EOS', 2], ['UNK', 3]]
        if included_prefix
        else []
    )
    count.extend(collections.Counter(words).most_common(n_words))
    dictionary = dict()
    for word, _ in count:
        dictionary[word] = len(dictionary)
    data = list()
    for word in words:
        index = dictionary.get(word, 3)
        data.append(index)
    reversed_dictionary = dict(zip(dictionary.values(), dictionary.keys()))
    return data, count, dictionary, reversed_dictionary


def features_crf(sentence, index):
    return {
        'word': sentence[index],
        'is_first': index == 0,
        'is_last': index == len(sentence) - 1,
        'prefix-1': sentence[index][0],
        'prefix-2': sentence[index][:2],
        'prefix-3': sentence[index][:3],
        'suffix-1': sentence[index][-1],
        'suffix-2': sentence[index][-2:],
        'suffix-3': sentence[index][-3:],
        'prev_word': '' if index == 0 else sentence[index - 1],
        'prev_word-prefix-1': '' if index == 0 else sentence[index - 1][0],
        'prev_word-prefix-2': '' if index == 0 else sentence[index - 1][:2],
        'prev_word-prefix-3': '' if index == 0 else sentence[index - 1][:3],
        'prev_word-suffix-1': '' if index == 0 else sentence[index - 1][-1],
        'prev_word-suffix-2': '' if index == 0 else sentence[index - 1][-2:],
        'prev_word-suffix-3': '' if index == 0 else sentence[index - 1][-3:],
        'next_word-prefix-1': ''
        if index == len(sentence) - 1
        else sentence[index + 1][0],
        'next_word-prefix-2': ''
        if index == len(sentence) - 1
        else sentence[index + 1][:2],
        'next_word-prefix-3': ''
        if index == len(sentence) - 1
        else sentence[index + 1][:3],
        'next_word-suffix-1': ''
        if index == len(sentence) - 1
        else sentence[index + 1][-1],
        'next_word-suffix-2': ''
        if index == len(sentence) - 1
        else sentence[index + 1][-2:],
        'next_word-suffix-3': ''
        if index == len(sentence) - 1
        else sentence[index + 1][-3:],
        'next_word': '' if index == len(sentence) - 1 else sentence[index + 1],
        'has_hyphen': '-' in sentence[index],
        'is_numeric': sentence[index].isdigit(),
    }


def most_common(l):
    return max(set(l), key = l.count)


def voting_stack(models, text):
    assert isinstance(models, list), 'models must be a list'
    assert isinstance(text, str), 'vectorizer must be a string'
    results, texts, votes = [], [], []
    for i in range(len(models)):
        assert 'predict' in dir(models[i]), 'all models must able to predict'
        predicted = np.array(models[i].predict(text))
        results.append(predicted[:, 1:2])
        texts.append(predicted[:, 0])
    concatenated = np.concatenate(results, axis = 1)
    for row in concatenated:
        votes.append(most_common(row.tolist()))
    return list(map(lambda X: (X[0], X[1]), list(zip(texts[-1], votes))))


def skipgrams(
    sequence,
    vocabulary_size,
    window_size = 4,
    negative_samples = 1.0,
    shuffle = True,
    categorical = False,
    sampling_table = None,
    seed = None,
):
    couples = []
    labels = []
    for i, wi in enumerate(sequence):
        if not wi:
            continue
        if sampling_table is not None:
            if sampling_table[wi] < random.random():
                continue

        window_start = max(0, i - window_size)
        window_end = min(len(sequence), i + window_size + 1)
        for j in range(window_start, window_end):
            if j != i:
                wj = sequence[j]
                if not wj:
                    continue
                couples.append([wi, wj])
                if categorical:
                    labels.append([0, 1])
                else:
                    labels.append(1)

    if negative_samples > 0:
        num_negative_samples = int(len(labels) * negative_samples)
        words = [c[0] for c in couples]
        random.shuffle(words)

        couples += [
            [words[i % len(words)], random.randint(1, vocabulary_size - 1)]
            for i in range(num_negative_samples)
        ]
        if categorical:
            labels += [[1, 0]] * num_negative_samples
        else:
            labels += [0] * num_negative_samples

    if shuffle:
        if seed is None:
            seed = random.randint(0, 10e6)
        random.seed(seed)
        random.shuffle(couples)
        random.seed(seed)
        random.shuffle(labels)

    return couples, labels



from . import home

MALAY_TEXT = home + '/malay-text.txt'

PATH_SUMMARIZE = {
    'model': home + '/summary_frozen_model.pb',
    'setting': home + '/dictionary-summary.json',
}

S3_PATH_SUMMARIZE = {
    'model': 'v7/summary/summary_frozen_model.pb',
    'setting': 'v7/summary/dictionary-summary.json',
}

PATH_TOXIC = {
    'multinomial': {
        'model': home + '/multinomials-toxic.pkl',
        'vector': home + '/vectorizer-toxic.pkl',
    },
    'logistic': {
        'model': home + '/logistics-toxic.pkl',
        'vector': home + '/vectorizer-toxic.pkl',
    },
    'luong': {
        'model': home + '/luong-toxic.pb',
        'setting': home + '/luong-toxic.json',
    },
    'bahdanau': {
        'model': home + '/bahdanau-toxic.pb',
        'setting': home + '/bahdanau-toxic.json',
    },
    'hierarchical': {
        'model': home + '/hierarchical-toxic.pb',
        'setting': home + '/hierarchical-toxic.json',
    },
    'fast-text': {
        'model': home + '/fasttext-toxic.pb',
        'setting': home + '/fasttext-toxic.json',
        'pickle': home + '/fasttext-toxic.pkl',
    },
    'entity-network': {
        'model': home + '/entity-toxic.pb',
        'setting': home + '/entity-toxic.json',
    },
}

S3_PATH_TOXIC = {
    'multinomial': {
        'model': 'v6/multinomials-toxic.pkl',
        'vector': 'v6/vectorizer-toxic.pkl',
    },
    'logistic': {
        'model': 'v6/logistics-toxic.pkl',
        'vector': 'v6/vectorizer-toxic.pkl',
    },
    'luong': {
        'model': 'v8/toxic/luong-toxic.pb',
        'setting': 'v8/toxic/luong-toxic.json',
    },
    'bahdanau': {
        'model': 'v8/toxic/bahdanau-toxic.pb',
        'setting': 'v8/toxic/bahdanau-toxic.json',
    },
    'hierarchical': {
        'model': 'v8/toxic/hierarchical-toxic.pb',
        'setting': 'v8/toxic/hierarchical-toxic.json',
    },
    'fast-text': {
        'model': 'v8/toxic/fasttext-toxic.pb',
        'setting': 'v8/toxic/fasttext-toxic.json',
        'pickle': 'v8/toxic/fasttext-toxic.pkl',
    },
    'entity-network': {
        'model': 'v8/toxic/entity-toxic.pb',
        'setting': 'v8/toxic/entity-toxic.json',
    },
}

PATH_POS = {
    'crf': {'model': home + '/crf-pos.pkl'},
    'xgb': {
        'model': home + '/xgb-pos.pkl',
        'vector': home + '/xgb-bow-pos.pkl',
    },
    'concat': {
        'model': home + '/concat-pos.pb',
        'setting': home + '/concat-pos.json',
    },
    'luong': {
        'model': home + '/luong-pos.pb',
        'setting': home + '/luong-pos.json',
    },
    'bahdanau': {
        'model': home + '/bahdanau-pos.pb',
        'setting': home + '/bahdanau-pos.json',
    },
    'entity-network': {
        'model': home + '/entity-pos.pb',
        'setting': home + '/entity-pos.json',
    },
}

S3_PATH_POS = {
    'crf': {'model': 'v8/pos/crf-pos.pkl'},
    'concat': {
        'model': 'v8/pos/concat-pos.pb',
        'setting': 'v8/pos/concat-pos.json',
    },
    'luong': {
        'model': 'v8/pos/luong-pos.pb',
        'setting': 'v8/pos/luong-pos.json',
    },
    'bahdanau': {
        'model': 'v8/pos/bahdanau-pos.pb',
        'setting': 'v8/pos/bahdanau-pos.json',
    },
    'entity-network': {
        'model': 'v8/pos/entity-pos.pb',
        'setting': 'v8/pos/entity-pos.json',
    },
}

PATH_NORMALIZER = {
    'deep': {
        'model': home + '/normalizer-deep.pb',
        'setting': home + '/normalizer-deep.json',
    }
}

S3_PATH_NORMALIZER = {
    'deep': {
        'model': 'v6/normalizer-deep.pb',
        'setting': 'v6/normalizer-deep.json',
    }
}

PATH_LANG_DETECTION = {
    'multinomial': {
        'model': home + '/multinomial-language-detection.pkl',
        'vector': home + '/vectorizer-language-detection.pkl',
    },
    'xgb': {
        'model': home + '/xgb-language-detection.pkl',
        'vector': home + '/vectorizer-language-detection.pkl',
    },
    'sgd': {
        'model': home + '/sgd-language-detection.pkl',
        'vector': home + '/vectorizer-language-detection.pkl',
    },
}

S3_PATH_LANG_DETECTION = {
    'multinomial': {
        'model': 'v8/language-detection/multinomial-language-detection.pkl',
        'vector': 'v8/language-detection/language-detection-vectorizer.pkl',
    },
    'xgb': {
        'model': 'v8/language-detection/xgboost-language-detection.pkl',
        'vector': 'v8/language-detection/language-detection-vectorizer.pkl',
    },
    'sgd': {
        'model': 'v8/language-detection/sgd-language-detection.pkl',
        'vector': 'v8/language-detection/language-detection-vectorizer.pkl',
    },
}

PATH_ENTITIES = {
    'crf': {'model': home + '/crf-entities.pkl'},
    'concat': {
        'model': home + '/concat-entities.pb',
        'setting': home + '/concat-entities.json',
    },
    'luong': {
        'model': home + '/luong-entities.pb',
        'setting': home + '/luong-entities.json',
    },
    'bahdanau': {
        'model': home + '/bahdanau-entities.pb',
        'setting': home + '/bahdanau-entities.json',
    },
    'entity-network': {
        'model': home + '/entity-entities.pb',
        'setting': home + '/entity-entities.json',
    },
}

S3_PATH_ENTITIES = {
    'crf': {'model': 'v8/entities/crf-entities.pkl'},
    'concat': {
        'model': 'v8/entities/concat-entities.pb',
        'setting': 'v8/entities/concat-entities.json',
    },
    'luong': {
        'model': 'v8/entities/luong-entities.pb',
        'setting': 'v8/entities/luong-entities.json',
    },
    'bahdanau': {
        'model': 'v8/entities/bahdanau-entities.pb',
        'setting': 'v8/entities/bahdanau-entities.json',
    },
    'entity-network': {
        'model': 'v8/entities/entity-entities.pb',
        'setting': 'v8/entities/entity-entities.json',
    },
}

PATH_SENTIMENTS = {
    'bert': {
        'model': home + '/bert-sentiment.pb',
        'setting': home + '/bert-sentiment.json',
    },
    'entity-network': {
        'model': home + '/entity-sentiment.pb',
        'setting': home + '/entity-sentiment.json',
    },
    'hierarchical': {
        'model': home + '/hierarchical-sentiment.pb',
        'setting': home + '/hierarchical-sentiment.json',
    },
    'bahdanau': {
        'model': home + '/bahdanau-sentiment.pb',
        'setting': home + '/bahdanau-sentiment.json',
    },
    'luong': {
        'model': home + '/luong-sentiment.pb',
        'setting': home + '/luong-sentiment.json',
    },
    'bidirectional': {
        'model': home + '/bidirectional-sentiment.pb',
        'setting': home + '/bidirectional-sentiment.json',
    },
    'fast-text': {
        'model': home + '/fasttext-sentiment.pb',
        'setting': home + '/fasttext-sentiment.json',
        'pickle': home + '/fasttext-sentiment.pkl',
    },
    'multinomial': {
        'model': home + '/multinomial-sentiment.pkl',
        'vector': home + '/multinomial-sentiment-tfidf.pkl',
    },
    'xgb': {
        'model': home + '/xgboost-sentiment.pkl',
        'vector': home + '/xgboost-sentiment-tfidf.pkl',
    },
}

S3_PATH_SENTIMENTS = {
    'bert': {
        'model': 'v8/sentiment/bert-sentiment.pb',
        'setting': 'v8/sentiment/bert-sentiment.json',
    },
    'entity-network': {
        'model': 'v8/sentiment/entity-sentiment.pb',
        'setting': 'v8/sentiment/entity-sentiment.json',
    },
    'hierarchical': {
        'model': 'v8/sentiment/hierarchical-sentiment.pb',
        'setting': 'v8/sentiment/hierarchical-sentiment.json',
    },
    'bahdanau': {
        'model': 'v8/sentiment/bahdanau-sentiment.pb',
        'setting': 'v8/sentiment/bahdanau-sentiment.json',
    },
    'luong': {
        'model': 'v8/sentiment/luong-sentiment.pb',
        'setting': 'v8/sentiment/luong-sentiment.json',
    },
    'bidirectional': {
        'model': 'v8/sentiment/bidirectional-sentiment.pb',
        'setting': 'v8/sentiment/bidirectional-sentiment.json',
    },
    'fast-text': {
        'model': 'v8/sentiment/fasttext-sentiment.pb',
        'setting': 'v8/sentiment/fasttext-sentiment.json',
        'pickle': 'v8/sentiment/fasttext-sentiment.pkl',
    },
    'multinomial': {
        'model': 'v8/sentiment/multinomial-sentiment.pkl',
        'vector': 'v8/sentiment/multinomial-sentiment-tfidf.pkl',
    },
    'xgb': {
        'model': 'v8/sentiment/xgboost-sentiment.pkl',
        'vector': 'v8/sentiment/xgboost-sentiment-tfidf.pkl',
    },
}



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import re
import os
import json
import tensorflow as tf
from unidecode import unidecode
from tensorflow.contrib.seq2seq.python.ops import beam_search_ops
from Sastrawi.Stemmer.StemmerFactory import StemmerFactory
from .tatabahasa import permulaan, hujung
from .utils import load_graph, download_file
from .text_functions import (
    pad_sentence_batch,
    stemmer_str_idx,
    classification_textcleaning,
)
from . import home

factory = StemmerFactory()
sastrawi = factory.create_stemmer()
stemmer_json = home + '/stemmer-deep.json'
stemmer_graph = home + '/stemmer-deep.pb'
GO = 0
PAD = 1
EOS = 2
UNK = 3


def classification_textcleaning_stemmer_attention(string):
    string = re.sub(
        'http\S+|www.\S+',
        '',
        ' '.join(
            [i for i in string.split() if i.find('#') < 0 and i.find('@') < 0]
        ),
    )
    string = unidecode(string).replace('.', ' . ').replace(',', ' , ')
    string = re.sub('[^A-Za-z ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string).strip()
    string = ' '.join(
        [i for i in re.findall('[\\w\']+|[;:\-\(\)&.,!?"]', string) if len(i)]
    )
    string = string.lower().split()
    string = [(naive_stemmer(word), word) for word in string]
    return (
        ' '.join([word[0] for word in string if len(word[0]) > 1]),
        ' '.join([word[1] for word in string if len(word[0]) > 1]),
    )


def classification_textcleaning_stemmer(string):
    string = re.sub(
        'http\S+|www.\S+',
        '',
        ' '.join(
            [i for i in string.split() if i.find('#') < 0 and i.find('@') < 0]
        ),
    )
    string = unidecode(string).replace('.', ' . ').replace(',', ' , ')
    string = re.sub('[^A-Za-z ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string).strip()
    string = ' '.join(
        [i for i in re.findall('[\\w\']+|[;:\-\(\)&.,!?"]', string) if len(i)]
    )
    string = string.lower().split()
    string = [(naive_stemmer(word), word) for word in string]
    return ' '.join([word[0] for word in string if len(word[0]) > 1])


class DEEP_STEMMER:
    def __init__(self, x, logits, sess, dicts):
        self._sess = sess
        self._x = x
        self._logits = logits
        self._dicts = dicts
        self._dicts['rev_dictionary_to'] = {
            int(k): v for k, v in self._dicts['rev_dictionary_to'].items()
        }

    def stem(self, string):
        """
        Stem a string.

        Parameters
        ----------
        string : str

        Returns
        -------
        string: stemmed string
        """
        assert isinstance(string, str), 'input must be a string'
        token_strings = classification_textcleaning(string, True).split()
        idx = stemmer_str_idx(token_strings, self._dicts['dictionary_from'])
        predicted = self._sess.run(
            self._logits, feed_dict = {self._x: pad_sentence_batch(idx, PAD)[0]}
        )
        results = []
        for word in predicted:
            results.append(
                ''.join(
                    [
                        self._dicts['rev_dictionary_to'][c]
                        for c in word
                        if c not in [GO, PAD, EOS, UNK]
                    ]
                )
            )
        return ' '.join(results)


def naive_stemmer(word):
    """
    Stem a string using Regex.

    Parameters
    ----------
    string : str

    Returns
    -------
    string: stemmed string
    """
    assert isinstance(word, str), 'input must be a string'
    hujung_result = re.findall(r'^(.*?)(%s)$' % ('|'.join(hujung)), word)
    word = hujung_result[0][0] if len(hujung_result) else word
    permulaan_result = re.findall(
        r'^(.*?)(%s)' % ('|'.join(permulaan[::-1])), word
    )
    permulaan_result.extend(
        re.findall(r'^(.*?)(%s)' % ('|'.join(permulaan)), word)
    )
    mula = permulaan_result if len(permulaan_result) else ''
    if len(mula):
        mula = mula[1][1] if len(mula[1][1]) > len(mula[0][1]) else mula[0][1]
    return word.replace(mula, '')


def sastrawi_stemmer(string):
    """
    Stem a string using Sastrawi.

    Parameters
    ----------
    string : str

    Returns
    -------
    string: stemmed string.
    """
    assert isinstance(string, str), 'input must be a string'
    return sastrawi.stem(string)


def deep_stemmer():
    """
    Load seq2seq stemmer deep learning model.

    Returns
    -------
    DEEP_STEMMER: malaya.stemmer.DEEP_STEMMER class
    """
    if not os.path.isfile(stemmer_json):
        print('downloading JSON stemmer')
        download_file('v5/stemmer-deep.json', stemmer_json)
    with open(stemmer_json, 'r') as fopen:
        dic_stemmer = json.load(fopen)
    if not os.path.isfile(stemmer_graph):
        print('downloading stemmer graph')
        download_file('v5/stemmer-frozen-model.pb', stemmer_graph)
    g = load_graph(stemmer_graph)
    return DEEP_STEMMER(
        g.get_tensor_by_name('import/Placeholder:0'),
        g.get_tensor_by_name('import/logits:0'),
        tf.InteractiveSession(graph = g),
        dic_stemmer,
    )



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import pickle
import os
from .utils import download_file
from .sklearn_model import LANGUAGE_DETECTION
from .paths import PATH_LANG_DETECTION, S3_PATH_LANG_DETECTION

lang_labels = {0: 'OTHER', 1: 'ENGLISH', 2: 'INDONESIA', 3: 'MALAY'}


def get_language_labels():
    """
    Return language labels dictionary.
    """
    return lang_labels


def multinomial_detect_languages():
    """
    Load multinomial language detection model.

    Returns
    -------
    LANGUAGE_DETECTION : malaya.sklearn_model.LANGUAGE_DETECTION class
    """
    if not os.path.isfile(PATH_LANG_DETECTION['multinomial']['vector']):
        print('downloading LANGUAGE-DETECTION pickled vectorizer')
        download_file(
            S3_PATH_LANG_DETECTION['multinomial']['vector'],
            PATH_LANG_DETECTION['multinomial']['vector'],
        )
    if not os.path.isfile(PATH_LANG_DETECTION['multinomial']['model']):
        print('downloading LANGUAGE-DETECTION pickled multinomial model')
        download_file(
            S3_PATH_LANG_DETECTION['multinomial']['model'],
            PATH_LANG_DETECTION['multinomial']['model'],
        )
    with open(PATH_LANG_DETECTION['multinomial']['vector'], 'rb') as fopen:
        vector = pickle.load(fopen)
    with open(PATH_LANG_DETECTION['multinomial']['model'], 'rb') as fopen:
        model = pickle.load(fopen)
    return LANGUAGE_DETECTION(model, lang_labels, vector)


def sgd_detect_languages():
    """
    Load SGD language detection model.

    Returns
    -------
    LANGUAGE_DETECTION : malaya.sklearn_model.LANGUAGE_DETECTION class
    """
    if not os.path.isfile(PATH_LANG_DETECTION['sgd']['vector']):
        print('downloading LANGUAGE-DETECTION pickled vectorizer')
        download_file(
            S3_PATH_LANG_DETECTION['sgd']['vector'],
            PATH_LANG_DETECTION['sgd']['vector'],
        )
    if not os.path.isfile(PATH_LANG_DETECTION['sgd']['model']):
        print('downloading LANGUAGE-DETECTION pickled SGD model')
        download_file(
            S3_PATH_LANG_DETECTION['sgd']['model'],
            PATH_LANG_DETECTION['sgd']['model'],
        )
    with open(PATH_LANG_DETECTION['sgd']['vector'], 'rb') as fopen:
        vector = pickle.load(fopen)
    with open(PATH_LANG_DETECTION['sgd']['model'], 'rb') as fopen:
        model = pickle.load(fopen)
    return LANGUAGE_DETECTION(model, lang_labels, vector)


def xgb_detect_languages():
    """
    Load XGB language detection model.

    Returns
    -------
    LANGUAGE_DETECTION : malaya.sklearn_model.LANGUAGE_DETECTION class
    """
    if not os.path.isfile(PATH_LANG_DETECTION['xgb']['vector']):
        print('downloading LANGUAGE-DETECTION pickled vectorizer')
        download_file(
            S3_PATH_LANG_DETECTION['xgb']['vector'],
            PATH_LANG_DETECTION['xgb']['vector'],
        )
    if not os.path.isfile(PATH_LANG_DETECTION['xgb']['model']):
        print('downloading LANGUAGE-DETECTION pickled XGB model')
        download_file(
            S3_PATH_LANG_DETECTION['xgb']['model'],
            PATH_LANG_DETECTION['xgb']['model'],
        )
    with open(PATH_LANG_DETECTION['xgb']['vector'], 'rb') as fopen:
        vector = pickle.load(fopen)
    with open(PATH_LANG_DETECTION['xgb']['model'], 'rb') as fopen:
        model = pickle.load(fopen)
    return LANGUAGE_DETECTION(model, lang_labels, vector, mode = 'xgb')



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import pickle
import os
import numpy as np
import tensorflow as tf
from fuzzywuzzy import fuzz
from sklearn.utils import shuffle
from sklearn.manifold import TSNE
from scipy.spatial.distance import cdist
from sklearn.neighbors import NearestNeighbors
from . import home
from .utils import download_file


def malaya_word2vec(size = 256):
    """
    Return malaya pretrained news word2vec

    Parameters
    ----------
    size: int, (default=256)

    Returns
    -------
    dictionary: dictionary of dictionary, reverse dictionary and vectors
    """
    assert isinstance(size, int), 'input must be an integer'
    if size not in [32, 64, 128, 256, 512]:
        raise Exception('size word2vec not supported')
    if not os.path.isfile('%s/word2vec-%d.p' % (home, size)):
        print('downloading word2vec-%d embedded' % (size))
        download_file(
            'v7/word2vec/word2vec-%d.p' % (size),
            '%s/word2vec-%d.p' % (home, size),
        )
    with open('%s/word2vec-%d.p' % (home, size), 'rb') as fopen:
        return pickle.load(fopen)


class Calculator:
    def __init__(self, tokens):
        self._tokens = tokens
        self._current = tokens[0]

    def exp(self):
        result = self.term()
        while self._current in ('+', '-'):
            if self._current == '+':
                self.next()
                result += self.term()
            if self._current == '-':
                self.next()
                result -= self.term()
        return result

    def factor(self):
        result = None
        if self._current[0].isdigit() or self._current[-1].isdigit():
            result = np.array([float(i) for i in self._current.split(',')])
            self.next()
        elif self._current is '(':
            self.next()
            result = self.exp()
            self.next()
        return result

    def next(self):
        self._tokens = self._tokens[1:]
        self._current = self._tokens[0] if len(self._tokens) > 0 else None

    def term(self):
        result = self.factor()
        while self._current in ('*', '/'):
            if self._current == '*':
                self.next()
                result *= self.term()
            if self._current == '/':
                self.next()
                result /= self.term()
        return result


class Word2Vec:
    def __init__(self, embed_matrix, dictionary):
        self._embed_matrix = embed_matrix
        self._dictionary = dictionary
        self._reverse_dictionary = {v: k for k, v in dictionary.items()}
        self.words = list(dictionary.keys())

    def get_vector_by_name(self, word):
        """
        get vector based on name

        Parameters
        ----------
        word: str

        Returns
        -------
        vector: numpy
        """
        return np.ravel(self._embed_matrix[self._dictionary[word], :])

    def calculator(
        self,
        equation,
        num_closest = 5,
        metric = 'cosine',
        return_similarity = True,
    ):
        """
        calculator parser for word2vec

        Parameters
        ----------
        equation: str
            Eg, '(mahathir + najib) - rosmah'
        num_closest: int, (default=5)
            number of words closest to the result
        metric: str, (default='cosine')
            vector distance algorithm
        return_similarity: bool, (default=True)
            if True, will return between 0-1 represents the distance

        Returns
        -------
        word_list: list of nearest words
        """
        assert isinstance(equation, str), 'input must be a string'
        assert isinstance(num_closest, int), 'num_closest must be an integer'
        assert isinstance(metric, str), 'metric must be a string'
        assert isinstance(
            return_similarity, bool
        ), 'num_closest must be a boolean'
        tokens, temp = [], ''
        for char in equation:
            if char == ' ':
                continue
            if char not in '()*+-':
                temp += char
            else:
                if len(temp):
                    row = self._dictionary[
                        self.words[
                            np.argmax([fuzz.ratio(temp, k) for k in self.words])
                        ]
                    ]
                    tokens.append(
                        ','.join(
                            self._embed_matrix[row, :].astype('str').tolist()
                        )
                    )
                    temp = ''
                tokens.append(char)
        if len(temp):
            row = self._dictionary[
                self.words[np.argmax([fuzz.ratio(temp, k) for k in self.words])]
            ]
            tokens.append(
                ','.join(self._embed_matrix[row, :].astype('str').tolist())
            )
        if return_similarity:
            nn = NearestNeighbors(num_closest + 1, metric = metric).fit(
                self._embed_matrix
            )
            distances, idx = nn.kneighbors(
                Calculator(tokens).exp().reshape((1, -1))
            )
            word_list = []
            for i in range(1, idx.shape[1]):
                word_list.append(
                    [self._reverse_dictionary[idx[0, i]], 1 - distances[0, i]]
                )
            return word_list
        else:
            closest_indices = self.closest_row_indices(
                Calculator(tokens).exp(), num_closest + 1, metric
            )
            word_list = []
            for i in closest_indices:
                word_list.append(self._reverse_dictionary[i])
            return word_list

    def n_closest(
        self, word, num_closest = 5, metric = 'cosine', return_similarity = True
    ):
        """
        find nearest words based on a word

        Parameters
        ----------
        word: str
            Eg, 'najib'
        num_closest: int, (default=5)
            number of words closest to the result
        metric: str, (default='cosine')
            vector distance algorithm
        return_similarity: bool, (default=True)
            if True, will return between 0-1 represents the distance

        Returns
        -------
        word_list: list of nearest words
        """
        assert isinstance(word, str), 'input must be a string'
        assert isinstance(num_closest, int), 'num_closest must be an integer'
        assert isinstance(metric, str), 'metric must be a string'
        assert isinstance(
            return_similarity, bool
        ), 'num_closest must be a boolean'
        if return_similarity:
            nn = NearestNeighbors(num_closest + 1, metric = metric).fit(
                self._embed_matrix
            )
            distances, idx = nn.kneighbors(
                self._embed_matrix[self._dictionary[word], :].reshape((1, -1))
            )
            word_list = []
            for i in range(1, idx.shape[1]):
                word_list.append(
                    [self._reverse_dictionary[idx[0, i]], 1 - distances[0, i]]
                )
            return word_list
        else:
            wv = self.get_vector_by_name(word)
            closest_indices = self.closest_row_indices(
                wv, num_closest + 1, metric
            )
            word_list = []
            for i in closest_indices:
                word_list.append(self._reverse_dictionary[i])
            if word in word_list:
                word_list.remove(word)
            return word_list

    def closest_row_indices(self, wv, num, metric):
        dist_array = np.ravel(
            cdist(self._embed_matrix, wv.reshape((1, -1)), metric = metric)
        )
        sorted_indices = np.argsort(dist_array)
        return sorted_indices[:num]

    def analogy(self, a, b, c, num = 1, metric = 'cosine'):
        """
        analogy calculation, vb - va + vc

        Parameters
        ----------
        a: str
        b: str
        c: str
        num: int, (default=1)
        metric: str, (default='cosine')
            vector distance algorithm

        Returns
        -------
        word_list: list of nearest words
        """
        assert isinstance(a, str), 'a must be a string'
        assert isinstance(b, str), 'b must be a string'
        assert isinstance(c, str), 'c must be a string'
        if a not in self._dictionary:
            raise Exception('a not in dictinary')
        if b not in self._dictionary:
            raise Exception('b not in dictinary')
        if c not in self._dictionary:
            raise Exception('c not in dictinary')
        va = self.get_vector_by_name(a)
        vb = self.get_vector_by_name(b)
        vc = self.get_vector_by_name(c)
        vd = vb - va + vc
        closest_indices = self.closest_row_indices(vd, num, metric)
        d_word_list = []
        for i in closest_indices:
            d_word_list.append(self._reverse_dictionary[i])
        return d_word_list

    def project_2d(self, start, end):
        """
        project word2vec into 2d dimension

        Parameters
        ----------
        start: int
        end: int

        Returns
        -------
        tsne decomposition: numpy
        """
        assert isinstance(start, int), 'start must be an integer'
        assert isinstance(end, int), 'end must be an integer'
        tsne = TSNE(n_components = 2)
        embed_2d = tsne.fit_transform(self._embed_matrix[start:end, :])
        word_list = []
        for i in range(start, end):
            word_list.append(self._reverse_dictionary[i])
        return embed_2d, word_list



import xgboost as xgb
import numpy as np
from collections import Counter
from scipy.sparse import hstack
from .text_functions import (
    simple_textcleaning,
    classification_textcleaning,
    features_crf,
    entities_textcleaning,
)

from collections import Counter


def transitions(trans_features):
    for (label_from, label_to), weight in trans_features:
        print('%-6s -> %-7s %0.6f' % (label_from, label_to, weight))


def state_features(state_features):
    for (attr, label), weight in state_features:
        print('%0.6f %-8s %s' % (weight, label, attr))


class CRF:
    def __init__(self, model, is_lower = True):
        self._model = model
        self._is_lower = is_lower

    def predict(self, string):
        """
        Tag a string

        Parameters
        ----------
        string : str

        Returns
        -------
        string: tagged string
        """
        assert isinstance(string, str), 'input must be a string'
        string = string.lower() if self._is_lower else string
        string = entities_textcleaning(string)
        batch_x = [features_crf(string, index) for index in range(len(string))]
        return [
            (string[no], tag)
            for no, tag in enumerate(self._model.predict_single(batch_x))
        ]

    def print_transitions(self, top_k = 10):
        """
        Print important top-k transitions

        Parameters
        ----------
        top_k : int
        """
        assert isinstance(top_k, int), 'input must be an integer'
        print('Top-%d likely transitions:' % (top_k))
        transitions(
            Counter(self._model.transition_features_).most_common(top_k)
        )

        print('\nTop-%d unlikely transitions:' % (top_k))
        transitions(
            Counter(self._model.transition_features_).most_common()[-top_k:]
        )

    def print_features(self, top_k = 10):
        """
        Print important top-k features

        Parameters
        ----------
        top_k : int
        """
        assert isinstance(top_k, int), 'input must be an integer'
        print('Top-%d positive:' % (top_k))
        state_features(Counter(self._model.state_features_).most_common(top_k))

        print('\nTop-%d negative:' % (top_k))
        state_features(
            Counter(self._model.state_features_).most_common()[-top_k:]
        )


class USER_XGB:
    def __init__(self, xgb, label, vectorize, cleaning = simple_textcleaning):
        self.xgb = xgb
        self.label = label
        self.vectorize = vectorize
        self._cleaning = cleaning

    def predict(self, string, get_proba = False):
        """
        Classify a string

        Parameters
        ----------
        string : str
        get_proba: bool, optional (default=False)
            If True, it will return probability of classes.

        Returns
        -------
        string: result
        """
        assert isinstance(string, str), 'input must be a string'
        vectors = self.vectorize.transform([self._cleaning(string)])
        result = self.xgb.predict(
            xgb.DMatrix(vectors), ntree_limit = self.xgb.best_ntree_limit
        )[0]
        if get_proba:
            return {self.label[i]: result[i] for i in range(len(result))}
        else:
            return self.label[np.argmax(result)]

    def predict_batch(self, strings, get_proba = False):
        """
        Classify a list of strings

        Parameters
        ----------
        strings: list
        get_proba: bool, optional (default=False)
            If True, it will return probability of classes.

        Returns
        -------
        string: list of results
        """
        assert isinstance(strings, list) and isinstance(
            strings[0], str
        ), 'input must be list of strings'
        strings = [self._cleaning(string) for string in strings]
        vectors = self.vectorize.transform(strings)
        results = self.xgb.predict(
            xgb.DMatrix(vectors), ntree_limit = self.xgb.best_ntree_limit
        )
        if get_proba:
            outputs = []
            for result in results:
                outputs.append(
                    {self.label[i]: result[i] for i in range(len(result))}
                )
            return outputs
        else:
            return [self.label[i] for i in np.argmax(results, axis = 1)]


class USER_BAYES:
    def __init__(
        self, multinomial, label, vectorize, cleaning = simple_textcleaning
    ):
        self.multinomial = multinomial
        self.label = label
        self.vectorize = vectorize
        self._cleaning = cleaning

    def predict(self, string, get_proba = False):
        """
        Classify a string

        Parameters
        ----------
        string : str
        get_proba: bool, optional (default=False)
            If True, it will return probability of classes.

        Returns
        -------
        string: result
        """
        assert isinstance(string, str), 'input must be a string'
        vectors = self.vectorize.transform([self._cleaning(string)])
        if get_proba:
            result = self.multinomial.predict_proba(vectors)[0]
            return {self.label[i]: result[i] for i in range(len(result))}
        else:
            return self.label[self.multinomial.predict(vectors)[0]]

    def predict_batch(self, strings, get_proba = False):
        """
        Classify a list of strings

        Parameters
        ----------
        strings: list
        get_proba: bool, optional (default=False)
            If True, it will return probability of classes.

        Returns
        -------
        string: list of results
        """
        assert isinstance(strings, list) and isinstance(
            strings[0], str
        ), 'input must be list of strings'
        strings = [self._cleaning(string) for string in strings]
        vectors = self.vectorize.transform(strings)
        if get_proba:
            results = self.multinomial.predict_proba(vectors)
            outputs = []
            for result in results:
                outputs.append(
                    {self.label[i]: result[i] for i in range(len(result))}
                )
            return outputs
        else:
            return [
                self.label[result]
                for result in self.multinomial.predict(vectors)
            ]


class TOXIC:
    def __init__(self, models, vectors):
        self._models = models
        self._vectors = vectors
        self._class_names = [
            'toxic',
            'severe_toxic',
            'obscene',
            'threat',
            'insult',
            'identity_hate',
        ]

    def _stack(self, strings):
        char_features = self._vectors['char'].transform(strings)
        word_features = self._vectors['word'].transform(strings)
        return hstack([char_features, word_features])

    def predict(self, string, get_proba = False):
        """
        Classify a string

        Parameters
        ----------
        string : str
        get_proba: bool, optional (default=False)
            If True, it will return probability of classes.

        Returns
        -------
        string: result
        """
        assert isinstance(string, str), 'input must be a string'
        stacked = self._stack([classification_textcleaning(string, True)])
        result = {}
        for no, label in enumerate(self._class_names):
            if get_proba:
                result[label] = self._models[no].predict_proba(stacked)[0, 1]
            else:
                result[label] = self._models[no].predict(stacked)[0]
        return result

    def predict_batch(self, strings, get_proba = False):
        """
        Classify a list of strings

        Parameters
        ----------
        strings: list
        get_proba: bool, optional (default=False)
            If True, it will return probability of classes.

        Returns
        -------
        string: list of results
        """
        assert isinstance(strings, list) and isinstance(
            strings[0], str
        ), 'input must be list of strings'
        stacked = self._stack(
            [classification_textcleaning(i, True) for i in strings]
        )
        result = {}
        for no, label in enumerate(self._class_names):
            if get_proba:
                result[label] = (
                    self._models[no].predict_proba(stacked)[:, 1].tolist()
                )
            else:
                result[label] = self._models[no].predict(stacked).tolist()
        return result


class LANGUAGE_DETECTION:
    def __init__(self, model, label, vectorizer, mode = 'sklearn'):
        self._model = model
        self._label = label
        self._vectorizer = vectorizer
        self._mode = mode

    def predict(self, string, get_proba = False):
        """
        Classify a string

        Parameters
        ----------
        string : str
        get_proba: bool, optional (default=False)
            If True, it will return probability of classes.

        Returns
        -------
        string: result
        """
        assert isinstance(string, str), 'input must be a string'
        string = string.lower()
        char_features = self._vectorizer['bow'].transform([string])
        tfidf_features = self._vectorizer['tfidf'].transform([string])
        vectors = hstack([char_features, tfidf_features])
        if self._mode == 'xgb':
            result = self._model.predict(
                xgb.DMatrix(vectors), ntree_limit = self._model.best_ntree_limit
            )[0]
            if get_proba:
                return {self._label[i]: result[i] for i in range(len(result))}
            else:
                return self._label[np.argmax(result)]
        else:
            if get_proba:
                result = self._model.predict_proba(vectors)[0]
                return {self._label[i]: result[i] for i in range(len(result))}
            else:
                return self._label[self._model.predict(vectors)[0]]

    def predict_batch(self, strings, get_proba = False):
        """
        Classify a list of strings

        Parameters
        ----------
        strings: list
        get_proba: bool, optional (default=False)
            If True, it will return probability of classes.

        Returns
        -------
        string: list of results
        """
        assert isinstance(strings, list) and isinstance(
            strings[0], str
        ), 'input must be list of strings'
        strings = [string.lower() for string in strings]
        char_features = self._vectorizer['bow'].transform(strings)
        tfidf_features = self._vectorizer['tfidf'].transform(strings)
        vectors = hstack([char_features, tfidf_features])

        if self._mode == 'xgb':
            results = self._model.predict(
                xgb.DMatrix(vectors), ntree_limit = self._model.best_ntree_limit
            )
            if get_proba:
                outputs = []
                for result in results:
                    outputs.append(
                        {self._label[i]: result[i] for i in range(len(result))}
                    )
                return outputs
            else:
                return [self._label[i] for i in np.argmax(results, axis = 1)]
        else:
            if get_proba:
                results = self._model.predict_proba(vectors)
                outputs = []
                for result in results:
                    outputs.append(
                        {self._label[i]: result[i] for i in range(len(result))}
                    )
                return outputs
            else:
                return [
                    self._label[result]
                    for result in self._model.predict(vectors)
                ]



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import os
import re
import pickle
import json
import tensorflow as tf
from .tatabahasa import tatabahasa_dict, hujung, permulaan
from .utils import download_file, load_graph
from .text_functions import entities_textcleaning
from .sklearn_model import CRF
from .tensorflow_model import TAGGING
from .paths import PATH_POS, S3_PATH_POS


def get_available_pos_models():
    """
    List available deep learning entities models, ['concat', 'bahdanau', 'luong','entity-network']
    """
    return ['concat', 'bahdanau', 'luong', 'entity-network']


def naive_POS_word(word):
    for key, vals in tatabahasa_dict.items():
        if word in vals:
            return (key, word)
    try:
        if len(re.findall(r'^(.*?)(%s)$' % ('|'.join(hujung[:1])), i)[0]) > 1:
            return ('KJ', word)
    except:
        pass
    try:
        if (
            len(re.findall(r'^(.*?)(%s)' % ('|'.join(permulaan[:-4])), word)[0])
            > 1
        ):
            return ('KJ', word)
    except Exception as e:
        pass
    if len(word) > 2:
        return ('KN', word)
    else:
        return ('', word)


def naive_pos(string):
    """
    Recognize POS in a string using Regex.

    Parameters
    ----------
    string: str

    Returns
    -------
    string : tokenized string with POS related
    """
    assert isinstance(string, str), 'input must be a string'
    string = string.lower()
    results = []
    for i in string.split():
        results.append(naive_POS_word(i))
    return results


def crf_pos():
    """
    Load CRF POS Recognition model.

    Returns
    -------
    CRF : malaya.sklearn_model.CRF class
    """
    if not os.path.isfile(PATH_POS['crf']['model']):
        print('downloading POS frozen CRF model')
        download_file(S3_PATH_POS['crf']['model'], PATH_POS['crf']['model'])
    with open(PATH_POS['crf']['model'], 'rb') as fopen:
        model = pickle.load(fopen)
    return CRF(model, is_lower = False)


def deep_pos(model = 'concat'):
    """
    Load deep learning POS Recognition model.

    Parameters
    ----------
    model : str, optional (default='bahdanau')
        Model architecture supported. Allowed values:

        * ``'concat'`` - Concating character and word embedded for BiLSTM
        * ``'bahdanau'`` - Concating character and word embedded including Bahdanau Attention for BiLSTM
        * ``'luong'`` - Concating character and word embedded including Luong Attention for BiLSTM
        * ``'entity-network'`` - Concating character and word embedded on hybrid Entity-Network and RNN

    Returns
    -------
    TAGGING: malaya.tensorflow_model.TAGGING class
    """
    assert isinstance(model, str), 'model must be a string'
    model = model.lower()
    if model in ['concat', 'bahdanau', 'luong', 'entity-network']:
        if not os.path.isfile(PATH_POS[model]['model']):
            print('downloading POS frozen %s model' % (model))
            download_file(S3_PATH_POS[model]['model'], PATH_POS[model]['model'])
        if not os.path.isfile(PATH_POS[model]['setting']):
            print('downloading POS %s dictionary' % (model))
            download_file(
                S3_PATH_POS[model]['setting'], PATH_POS[model]['setting']
            )
        with open(PATH_POS[model]['setting'], 'r') as fopen:
            nodes = json.loads(fopen.read())
        g = load_graph(PATH_POS[model]['model'])
        if model == 'entity-network':
            return TAGGING(
                g.get_tensor_by_name('import/question:0'),
                g.get_tensor_by_name('import/char_ids:0'),
                g.get_tensor_by_name('import/logits:0'),
                nodes,
                tf.InteractiveSession(graph = g),
                model,
                is_lower = False,
                story = g.get_tensor_by_name('import/story:0'),
            )
        else:
            return TAGGING(
                g.get_tensor_by_name('import/Placeholder:0'),
                g.get_tensor_by_name('import/Placeholder_1:0'),
                g.get_tensor_by_name('import/logits:0'),
                nodes,
                tf.InteractiveSession(graph = g),
                model,
                is_lower = False,
            )
    else:
        raise Exception(
            'model not supported, please check supported models from malaya.get_available_pos_models()'
        )



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

from itertools import combinations
from toolz import compose
from sklearn.feature_extraction.text import CountVectorizer


class SkipGramVectorizer(CountVectorizer):
    def __init__(self, skip = 1, **kwds):
        super(SkipGramVectorizer, self).__init__(**kwds)
        self.skip = skip

    def build_sent_analyzer(self, preprocess, stop_words, tokenize):
        return lambda sent: self._word_skip_grams(
            compose(tokenize, preprocess, self.decode)(sent), stop_words
        )

    def build_analyzer(self):
        preprocess = self.build_preprocessor()
        stop_words = self.get_stop_words()
        tokenize = self.build_tokenizer()
        sent_analyze = self.build_sent_analyzer(
            preprocess, stop_words, tokenize
        )

        return lambda doc: self._sent_skip_grams(doc, sent_analyze)

    def _sent_skip_grams(self, doc, sent_analyze):
        skip_grams = []
        for sent in doc.split('\n'):
            skip_grams.extend(sent_analyze(sent))
        return skip_grams

    def _word_skip_grams(self, tokens, stop_words = None):
        if stop_words is not None:
            tokens = [w for w in tokens if w not in stop_words]
        min_n, max_n = self.ngram_range
        skip = self.skip
        if max_n != 1:
            original_tokens = tokens
            if min_n == 1:
                tokens = list(original_tokens)
                min_n += 1
            else:
                tokens = []

            n_original_tokens = len(original_tokens)
            tokens_append = tokens.append
            space_join = ' '.join

            for n in range(min_n, min(max_n + 1, n_original_tokens + 1)):
                for i in range(n_original_tokens - n + 1):
                    head = [original_tokens[i]]
                    for skip_tail in combinations(
                        original_tokens[i + 1 : i + n + skip], n - 1
                    ):
                        tokens_append(space_join(head + list(skip_tail)))
        return tokens



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

from sklearn import metrics, datasets
import tensorflow as tf
import numpy as np
import os
import json
import pickle
import random
from sklearn.naive_bayes import MultinomialNB
from sklearn.preprocessing import LabelEncoder
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer
from sklearn.cross_validation import train_test_split
from .text_functions import separate_dataset
from .stemmer import classification_textcleaning_stemmer
from .sklearn_model import USER_XGB, USER_BAYES
from .utils import download_file, load_graph
from .tensorflow_model import SENTIMENT
from .paths import PATH_SENTIMENTS, S3_PATH_SENTIMENTS
from .vectorizer import SkipGramVectorizer


def get_available_sentiment_models():
    """
    List available deep learning sentiment analysis models.
    """
    return [
        'fast-text',
        'hierarchical',
        'bahdanau',
        'luong',
        'bidirectional',
        'bert',
        'entity-network',
    ]


def deep_sentiment(model = 'luong'):
    """
    Load deep learning sentiment analysis model.

    Parameters
    ----------
    model : str, optional (default='luong')
        Model architecture supported. Allowed values:

        * ``'fast-text'`` - Fast-text architecture, embedded and logits layers only
        * ``'hierarchical'`` - LSTM with hierarchical attention architecture
        * ``'bahdanau'`` - LSTM with bahdanau attention architecture
        * ``'bidirectional'`` - LSTM with Bidirectional RNN architecture
        * ``'luong'`` - LSTM with luong attention architecture
        * ``'bert'`` - Deep Bidirectional transformers architecture
        * ``'entity-network'`` - Recurrent Entity-Network architecture

    Returns
    -------
    SENTIMENT: malaya.tensorflow_model.SENTIMENT class
    """
    assert isinstance(model, str), 'model must be a string'
    model = model.lower()
    if model == 'fast-text':
        if not os.path.isfile(PATH_SENTIMENTS['fast-text']['model']):
            print('downloading SENTIMENT frozen fast-text model')
            download_file(
                S3_PATH_SENTIMENTS['fast-text']['model'],
                PATH_SENTIMENTS['fast-text']['model'],
            )
        if not os.path.isfile(PATH_SENTIMENTS['fast-text']['setting']):
            print('downloading SENTIMENT fast-text dictionary')
            download_file(
                S3_PATH_SENTIMENTS['fast-text']['setting'],
                PATH_SENTIMENTS['fast-text']['setting'],
            )
        if not os.path.isfile(PATH_SENTIMENTS['fast-text']['pickle']):
            print('downloading SENTIMENT fast-text bigrams')
            download_file(
                S3_PATH_SENTIMENTS['fast-text']['pickle'],
                PATH_SENTIMENTS['fast-text']['pickle'],
            )
        with open(PATH_SENTIMENTS['fast-text']['setting'], 'r') as fopen:
            dictionary = json.load(fopen)['dictionary']
        with open(PATH_SENTIMENTS['fast-text']['pickle'], 'rb') as fopen:
            ngram = pickle.load(fopen)
        g = load_graph(PATH_SENTIMENTS['fast-text']['model'])
        return SENTIMENT(
            g.get_tensor_by_name('import/Placeholder:0'),
            g.get_tensor_by_name('import/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            ngram = ngram,
        )
    elif model == 'hierarchical':
        if not os.path.isfile(PATH_SENTIMENTS['hierarchical']['model']):
            print('downloading SENTIMENT frozen hierarchical model')
            download_file(
                S3_PATH_SENTIMENTS['hierarchical']['model'],
                PATH_SENTIMENTS['hierarchical']['model'],
            )
        if not os.path.isfile(PATH_SENTIMENTS['hierarchical']['setting']):
            print('downloading SENTIMENT hierarchical dictionary')
            download_file(
                S3_PATH_SENTIMENTS['hierarchical']['setting'],
                PATH_SENTIMENTS['hierarchical']['setting'],
            )
        with open(PATH_SENTIMENTS['hierarchical']['setting'], 'r') as fopen:
            dictionary = json.load(fopen)['dictionary']
        g = load_graph(PATH_SENTIMENTS['hierarchical']['model'])
        return SENTIMENT(
            g.get_tensor_by_name('import/Placeholder:0'),
            g.get_tensor_by_name('import/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            alphas = g.get_tensor_by_name('import/alphas:0'),
        )
    elif model == 'bahdanau':
        if not os.path.isfile(PATH_SENTIMENTS['bahdanau']['model']):
            print('downloading SENTIMENT frozen bahdanau model')
            download_file(
                S3_PATH_SENTIMENTS['bahdanau']['model'],
                PATH_SENTIMENTS['bahdanau']['model'],
            )
        if not os.path.isfile(PATH_SENTIMENTS['bahdanau']['setting']):
            print('downloading SENTIMENT bahdanau dictionary')
            download_file(
                S3_PATH_SENTIMENTS['bahdanau']['setting'],
                PATH_SENTIMENTS['bahdanau']['setting'],
            )
        with open(PATH_SENTIMENTS['bahdanau']['setting'], 'r') as fopen:
            dictionary = json.load(fopen)['dictionary']
        g = load_graph(PATH_SENTIMENTS['bahdanau']['model'])
        return SENTIMENT(
            g.get_tensor_by_name('import/Placeholder:0'),
            g.get_tensor_by_name('import/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            alphas = g.get_tensor_by_name('import/alphas:0'),
        )
    elif model == 'luong':
        if not os.path.isfile(PATH_SENTIMENTS['luong']['model']):
            print('downloading SENTIMENT frozen luong model')
            download_file(
                S3_PATH_SENTIMENTS['luong']['model'],
                PATH_SENTIMENTS['luong']['model'],
            )
        if not os.path.isfile(PATH_SENTIMENTS['luong']['setting']):
            print('downloading SENTIMENT luong dictionary')
            download_file(
                S3_PATH_SENTIMENTS['luong']['setting'],
                PATH_SENTIMENTS['luong']['setting'],
            )
        with open(PATH_SENTIMENTS['luong']['setting'], 'r') as fopen:
            dictionary = json.load(fopen)
        g = load_graph(PATH_SENTIMENTS['luong']['model'])
        return SENTIMENT(
            g.get_tensor_by_name('import/Placeholder:0'),
            g.get_tensor_by_name('import/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            alphas = g.get_tensor_by_name('import/alphas:0'),
        )
    elif model == 'bidirectional':
        if not os.path.isfile(PATH_SENTIMENTS['bidirectional']['model']):
            print('downloading SENTIMENT frozen bidirectional model')
            download_file(
                S3_PATH_SENTIMENTS['bidirectional']['model'],
                PATH_SENTIMENTS['bidirectional']['model'],
            )
        if not os.path.isfile(PATH_SENTIMENTS['bidirectional']['setting']):
            print('downloading SENTIMENT bidirectional dictionary')
            download_file(
                S3_PATH_SENTIMENTS['bidirectional']['setting'],
                PATH_SENTIMENTS['bidirectional']['setting'],
            )
        with open(PATH_SENTIMENTS['bidirectional']['setting'], 'r') as fopen:
            dictionary = json.load(fopen)
        g = load_graph(PATH_SENTIMENTS['bidirectional']['model'])
        return SENTIMENT(
            g.get_tensor_by_name('import/Placeholder:0'),
            g.get_tensor_by_name('import/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
        )
    elif model == 'bert':
        if not os.path.isfile(PATH_SENTIMENTS['bert']['model']):
            print('downloading SENTIMENT frozen bert model')
            download_file(
                S3_PATH_SENTIMENTS['bert']['model'],
                PATH_SENTIMENTS['bert']['model'],
            )
        if not os.path.isfile(PATH_SENTIMENTS['bert']['setting']):
            print('downloading SENTIMENT bert dictionary')
            download_file(
                S3_PATH_SENTIMENTS['bert']['setting'],
                PATH_SENTIMENTS['bert']['setting'],
            )
        with open(PATH_SENTIMENTS['bert']['setting'], 'r') as fopen:
            dictionary = json.load(fopen)
        g = load_graph(PATH_SENTIMENTS['bert']['model'])
        return SENTIMENT(
            g.get_tensor_by_name('import/Placeholder_input_ids:0'),
            g.get_tensor_by_name('import/loss/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            input_mask = g.get_tensor_by_name(
                'import/Placeholder_input_mask:0'
            ),
            segment_ids = g.get_tensor_by_name(
                'import/Placeholder_segment_ids:0'
            ),
            is_training = g.get_tensor_by_name(
                'import/Placeholder_is_training:0'
            ),
        )
    elif model == 'entity-network':
        if not os.path.isfile(PATH_SENTIMENTS['entity-network']['model']):
            print('downloading SENTIMENT frozen entity-network model')
            download_file(
                S3_PATH_SENTIMENTS['entity-network']['model'],
                PATH_SENTIMENTS['entity-network']['model'],
            )
        if not os.path.isfile(PATH_SENTIMENTS['entity-network']['setting']):
            print('downloading SENTIMENT entity-network dictionary')
            download_file(
                S3_PATH_SENTIMENTS['entity-network']['setting'],
                PATH_SENTIMENTS['entity-network']['setting'],
            )
        with open(PATH_SENTIMENTS['entity-network']['setting'], 'r') as fopen:
            dictionary = json.load(fopen)
        g = load_graph(PATH_SENTIMENTS['entity-network']['model'])
        return SENTIMENT(
            g.get_tensor_by_name('import/Placeholder_question:0'),
            g.get_tensor_by_name('import/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            dropout_keep_prob = g.get_tensor_by_name(
                'import/Placeholder_dropout_keep_prob:0'
            ),
            story = g.get_tensor_by_name('import/Placeholder_story:0'),
        )
    else:
        raise Exception(
            'model sentiment not supported, please check supported models from malaya.get_available_sentiment_models()'
        )


def bayes_sentiment(corpus, vector = 'tfidf', split_size = 0.2, **kwargs):
    """
    Train a multinomial model using your own corpus / dataset.

    Parameters
    ----------
    corpus : list, str
        List of (string, label). Allowed tuple and list only.

    vector: str
        Vectorization techniques to feed into the model. Allowed values:

        * ``'tfidf'`` - TF-IDF
        * ``'bow'`` - bag-of-word
        * ``'skip-gram'`` - bag-of-word with skipping module

    split_size: float
        Split size to train and test. Should bigger than 0, less than 1.

    Returns
    -------
    USER_BAYES: malaya.sklearn_model.USER_BAYES class
    """
    assert (
        isinstance(corpus, str)
        or isinstance(corpus, list)
        or isinstance(corpus, tuple)
    ), 'corpus must be a string location or list of strings or tuple of strings'
    assert isinstance(vector, str), 'vector must be a string'
    assert isinstance(split_size, float), 'split_size must be a float'
    assert (
        split_size > 0 and split_size < 1
    ), 'split_size must bigger than 0, less than 1'
    multinomial, labels, vectorize = None, None, None
    if isinstance(corpus, str):
        trainset = datasets.load_files(
            container_path = corpus, encoding = 'UTF-8'
        )
        trainset.data, trainset.target = separate_dataset(trainset)
        data, target = trainset.data, trainset.target
        labels = trainset.target_names
    if isinstance(corpus, list) or isinstance(corpus, tuple):
        assert (
            len(corpus[0]) == 2
        ), 'element of corpus must be list or tuple of (string, label)'
        assert isinstance(
            corpus[0][0], str
        ), 'left hand side of element must be a string'
        corpus = np.array(corpus)
        data, target = corpus[:, 0].tolist(), corpus[:, 1].tolist()
        labels = np.unique(target).tolist()
        target = LabelEncoder().fit_transform(target)
    c = list(zip(data, target))
    random.shuffle(c)
    data, target = zip(*c)
    data, target = list(data), list(target)
    for i in range(len(data)):
        data[i] = classification_textcleaning_stemmer(data[i])
    if 'tfidf' in vector.lower():
        vectorize = TfidfVectorizer(**kwargs).fit(data)
        vectors = vectorize.transform(data)
    elif 'bow' in vector.lower():
        vectorize = CountVectorizer(**kwargs).fit(data)
        vectors = vectorize.transform(data)
    elif 'skip-gram' in vector.lower():
        vectorize = SkipGramVectorizer(**kwargs).fit(data)
        vectors = vectorize.transform(data)
    else:
        raise Exception(
            "vectorizing techniques not supported, only support ['tf-idf', 'bow', 'skip-gram']"
        )
    multinomial = MultinomialNB()
    if split_size:
        train_X, test_X, train_Y, test_Y = train_test_split(
            vectors, target, test_size = split_size
        )
        multinomial.partial_fit(train_X, train_Y, classes = np.unique(target))
        predicted = multinomial.predict(test_X)
        print(
            metrics.classification_report(
                test_Y, predicted, target_names = labels
            )
        )
    else:
        multinomial.partial_fit(vectors, target, classes = np.unique(target))
        predicted = multinomial.predict(vectors)
        print(
            metrics.classification_report(
                target, predicted, target_names = labels
            )
        )
    return USER_BAYES(
        multinomial,
        labels,
        vectorize,
        cleaning = classification_textcleaning_stemmer,
    )


def pretrained_bayes_sentiment():
    """
    Load multinomial sentiment model.

    Returns
    -------
    USER_BAYES : malaya.sklearn_model.USER_BAYES class
    """
    if not os.path.isfile(PATH_SENTIMENTS['multinomial']['model']):
        print('downloading SENTIMENT pickled multinomial model')
        download_file(
            S3_PATH_SENTIMENTS['multinomial']['model'],
            PATH_SENTIMENTS['multinomial']['model'],
        )
    if not os.path.isfile(PATH_SENTIMENTS['multinomial']['vector']):
        print('downloading SENTIMENT pickled multinomial tfidf vectorization')
        download_file(
            S3_PATH_SENTIMENTS['multinomial']['vector'],
            PATH_SENTIMENTS['multinomial']['vector'],
        )
    with open(PATH_SENTIMENTS['multinomial']['model'], 'rb') as fopen:
        multinomial = pickle.load(fopen)
    with open(PATH_SENTIMENTS['multinomial']['vector'], 'rb') as fopen:
        vectorize = pickle.load(fopen)
    return USER_BAYES(
        multinomial,
        ['negative', 'positive'],
        vectorize,
        cleaning = classification_textcleaning_stemmer,
    )


def pretrained_xgb_sentiment():
    """
    Load XGB sentiment model.

    Returns
    -------
    USER_BAYES : malaya.sklearn_model.USER_BAYES class
    """
    if not os.path.isfile(PATH_SENTIMENTS['xgb']['model']):
        print('downloading SENTIMENT pickled XGB model')
        download_file(
            S3_PATH_SENTIMENTS['xgb']['model'], PATH_SENTIMENTS['xgb']['model']
        )
    if not os.path.isfile(PATH_SENTIMENTS['xgb']['vector']):
        print('downloading SENTIMENT pickled XGB tfidf vectorization')
        download_file(
            S3_PATH_SENTIMENTS['xgb']['vector'],
            PATH_SENTIMENTS['xgb']['vector'],
        )
    with open(PATH_SENTIMENTS['xgb']['model'], 'rb') as fopen:
        xgb = pickle.load(fopen)
    with open(PATH_SENTIMENTS['xgb']['vector'], 'rb') as fopen:
        vectorize = pickle.load(fopen)
    return USER_XGB(
        xgb,
        ['negative', 'positive'],
        vectorize,
        cleaning = classification_textcleaning_stemmer,
    )



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import tensorflow as tf
import numpy as np
from .text_functions import (
    str_idx,
    add_ngram,
    fasttext_str_idx,
    entities_textcleaning,
    char_str_idx,
    generate_char_seq,
)
from .stemmer import classification_textcleaning_stemmer_attention


class TAGGING:
    def __init__(
        self,
        X,
        X_char,
        logits,
        settings,
        sess,
        model,
        is_lower = True,
        story = None,
    ):
        self._X = X
        self._X_char = X_char
        self._logits = logits
        self._settings = settings
        self._sess = sess
        self._model = model
        self._is_lower = is_lower
        self._story = story

        self._settings['idx2word'] = {
            int(k): v for k, v in self._settings['idx2word'].items()
        }
        self._settings['idx2tag'] = {
            int(k): v for k, v in self._settings['idx2tag'].items()
        }

    def predict(self, string):
        """
        Tagging a string.

        Parameters
        ----------
        string : str

        Returns
        -------
        string: tagged string
        """
        assert isinstance(string, str), 'input must be a string'
        string = string.lower() if self._is_lower else string
        string = entities_textcleaning(string)
        batch_x = char_str_idx([string], self._settings['word2idx'], 2)
        batch_x_char = generate_char_seq(
            batch_x, self._settings['idx2word'], self._settings['char2idx']
        )
        if self._model == 'entity-network':
            batch_x_expand = np.expand_dims(batch_x, axis = 1)
            predicted = self._sess.run(
                self._logits,
                feed_dict = {
                    self._X: batch_x,
                    self._X_char: batch_x_char,
                    self._story: batch_x_expand,
                },
            )[0]
        else:
            predicted = self._sess.run(
                self._logits,
                feed_dict = {self._X: batch_x, self._X_char: batch_x_char},
            )[0]
        return [
            (string[i], self._settings['idx2tag'][predicted[i]])
            for i in range(len(predicted))
        ]


class SENTIMENT:
    def __init__(
        self,
        X,
        logits,
        sess,
        mode,
        dictionary,
        ngram = None,
        alphas = None,
        input_mask = None,
        segment_ids = None,
        is_training = None,
        dropout_keep_prob = None,
        story = None,
        maxlen = 80,
    ):
        self._X = X
        self._logits = logits
        self._sess = sess
        self._mode = mode
        self._dictionary = dictionary
        self._ngram = ngram
        self._alphas = alphas
        self._input_mask = input_mask
        self._segment_ids = segment_ids
        self._is_training = is_training
        self._dropout_keep_prob = dropout_keep_prob
        self._story = story
        self._maxlen = maxlen

    def predict(self, string):
        """
        classify a string.

        Parameters
        ----------
        string : str

        Returns
        -------
        dictionary: results
        """
        assert isinstance(string, str), 'input must be a string'
        string = classification_textcleaning_stemmer_attention(string)
        splitted = string[1].split()
        if self._mode == 'fast-text':
            batch_x = fasttext_str_idx([string[0]], self._dictionary)
            batch_x = add_ngram(batch_x, self._ngram)
        else:
            if self._mode in ['entity-network', 'bert']:
                batch_x = str_idx(
                    [string[0]], self._dictionary, self._maxlen, UNK = 3
                )
            else:
                batch_x = str_idx(
                    [string[0]], self._dictionary, len(splitted), UNK = 3
                )
        if self._mode in ['luong', 'bahdanau', 'hierarchical']:
            probs, alphas = self._sess.run(
                [tf.nn.softmax(self._logits), self._alphas],
                feed_dict = {self._X: batch_x},
            )
            if self._mode == 'hierarchical':
                alphas = alphas[0]
            words = []
            for i in range(alphas.shape[0]):
                words.append([splitted[i], alphas[i]])
            return {
                'negative': probs[0, 0],
                'positive': probs[0, 1],
                'attention': words,
            }
        if self._mode in ['bidirectional', 'fast-text']:
            probs = self._sess.run(
                tf.nn.softmax(self._logits), feed_dict = {self._X: batch_x}
            )
            return {'negative': probs[0, 0], 'positive': probs[0, 1]}
        if self._mode == 'bert':
            np_mask = np.ones((1, self._maxlen), dtype = np.int32)
            np_segment = np.ones((1, self._maxlen), dtype = np.int32)
            probs = self._sess.run(
                tf.nn.softmax(self._logits),
                feed_dict = {
                    self._X: batch_x,
                    self._input_mask: np_mask,
                    self._segment_ids: np_segment,
                    self._is_training: False,
                },
            )
            return {'negative': probs[0, 0], 'positive': probs[0, 1]}
        if self._mode == 'entity-network':
            batch_x_expand = np.expand_dims(batch_x, axis = 1)
            probs = self._sess.run(
                tf.nn.softmax(self._logits),
                feed_dict = {
                    self._X: batch_x,
                    self._story: batch_x_expand,
                    self._dropout_keep_prob: 1.0,
                },
            )
            return {'negative': probs[0, 0], 'positive': probs[0, 1]}

    def predict_batch(self, strings):
        """
        classify list of strings

        Parameters
        ----------
        strings : list

        Returns
        -------
        list_dictionaries: list of results
        """
        assert isinstance(strings, list) and isinstance(
            strings[0], str
        ), 'input must be list of strings'
        strings = [
            classification_textcleaning_stemmer_attention(i)[0] for i in strings
        ]
        maxlen = max([len(i.split()) for i in strings])
        if self._mode == 'fast-text':
            batch_x = fasttext_str_idx(strings, self._dictionary)
            batch_x = add_ngram(batch_x, self._ngram)
            batch_x = tf.keras.preprocessing.sequence.pad_sequences(
                batch_x, maxlen
            )
        else:
            if self._mode in ['entity-network', 'bert']:
                batch_x = str_idx(
                    strings, self._dictionary, self._maxlen, UNK = 3
                )
            else:
                batch_x = str_idx(strings, self._dictionary, maxlen, UNK = 3)
        if self._mode not in ['bert', 'entity-network']:
            probs = self._sess.run(
                tf.nn.softmax(self._logits), feed_dict = {self._X: batch_x}
            )
        if self._mode == 'bert':
            np_mask = np.ones((len(batch_x), self._maxlen), dtype = np.int32)
            np_segment = np.ones((len(batch_x), self._maxlen), dtype = np.int32)
            probs = self._sess.run(
                tf.nn.softmax(self._logits),
                feed_dict = {
                    self._X: batch_x,
                    self._input_mask: np_mask,
                    self._segment_ids: np_segment,
                    self._is_training: False,
                },
            )
        if self._mode == 'entity-network':
            batch_x_expand = np.expand_dims(batch_x, axis = 1)
            probs = self._sess.run(
                tf.nn.softmax(self._logits),
                feed_dict = {
                    self._X: batch_x,
                    self._story: batch_x_expand,
                    self._dropout_keep_prob: 1.0,
                },
            )

        dicts = []
        for i in range(probs.shape[0]):
            dicts.append({'negative': probs[i, 0], 'positive': probs[i, 1]})
        return dicts


class DEEP_TOXIC:
    def __init__(
        self,
        X,
        logits,
        sess,
        mode,
        dictionary,
        ngram = None,
        alphas = None,
        input_mask = None,
        segment_ids = None,
        is_training = None,
        dropout_keep_prob = None,
        story = None,
        maxlen = 80,
    ):
        self._X = X
        self._logits = logits
        self._sess = sess
        self._mode = mode
        self._dictionary = dictionary
        self._ngram = ngram
        self._alphas = alphas
        self._input_mask = input_mask
        self._segment_ids = segment_ids
        self._is_training = is_training
        self._dropout_keep_prob = dropout_keep_prob
        self._story = story
        self._maxlen = maxlen
        self._label = [
            'toxic',
            'severe_toxic',
            'obscene',
            'threat',
            'insult',
            'identity_hate',
        ]

    def predict(self, string):
        """
        classify a string.

        Parameters
        ----------
        string : str

        Returns
        -------
        dictionary: results
        """
        assert isinstance(string, str), 'input must be a string'
        string = classification_textcleaning_stemmer_attention(string)
        splitted = string[1].split()
        if self._mode == 'fast-text':
            batch_x = fasttext_str_idx([string[0]], self._dictionary)
            batch_x = add_ngram(batch_x, self._ngram)
        else:
            if self._mode in ['entity-network', 'bert']:
                batch_x = str_idx(
                    [string[0]], self._dictionary, self._maxlen, UNK = 3
                )
            else:
                batch_x = str_idx(
                    [string[0]], self._dictionary, len(splitted), UNK = 3
                )
        if self._mode in ['luong', 'bahdanau', 'hierarchical']:
            probs, alphas = self._sess.run(
                [tf.nn.sigmoid(self._logits), self._alphas],
                feed_dict = {self._X: batch_x},
            )
            if self._mode == 'hierarchical':
                alphas = alphas[0]
            words = []
            for i in range(alphas.shape[0]):
                words.append([splitted[i], alphas[i]])
            dict_result = {}
            for no in range(len(self._label)):
                dict_result[self._label[no]] = probs[0, no]
            dict_result['attention'] = words
            return dict_result
        if self._mode in ['fast-text']:
            probs = self._sess.run(
                tf.nn.softmax(self._logits), feed_dict = {self._X: batch_x}
            )
            dict_result = {}
            for no in range(len(self._label)):
                dict_result[self._label[no]] = probs[0, no]
            return dict_result
        if self._mode == 'entity-network':
            batch_x_expand = np.expand_dims(batch_x, axis = 1)
            probs = self._sess.run(
                tf.nn.softmax(self._logits),
                feed_dict = {
                    self._X: batch_x,
                    self._story: batch_x_expand,
                    self._dropout_keep_prob: 1.0,
                },
            )
            dict_result = {}
            for no in range(len(self._label)):
                dict_result[self._label[no]] = probs[0, no]
            return dict_result

    def predict_batch(self, strings):
        """
        classify list of strings

        Parameters
        ----------
        strings : list

        Returns
        -------
        list_dictionaries: list of results
        """
        assert isinstance(strings, list) and isinstance(
            strings[0], str
        ), 'input must be list of strings'
        strings = [
            classification_textcleaning_stemmer_attention(i)[0] for i in strings
        ]
        maxlen = max([len(i.split()) for i in strings])
        if self._mode == 'fast-text':
            batch_x = fasttext_str_idx(strings, self._dictionary)
            batch_x = add_ngram(batch_x, self._ngram)
            batch_x = tf.keras.preprocessing.sequence.pad_sequences(
                batch_x, maxlen
            )
        else:
            if self._mode in ['entity-network']:
                batch_x = str_idx(
                    strings, self._dictionary, self._maxlen, UNK = 3
                )
            else:
                batch_x = str_idx(strings, self._dictionary, maxlen, UNK = 3)
        if self._mode not in ['entity-network']:
            probs = self._sess.run(
                tf.nn.sigmoid(self._logits), feed_dict = {self._X: batch_x}
            )
        if self._mode == 'entity-network':
            batch_x_expand = np.expand_dims(batch_x, axis = 1)
            probs = self._sess.run(
                tf.nn.sigmoid(self._logits),
                feed_dict = {
                    self._X: batch_x,
                    self._story: batch_x_expand,
                    self._dropout_keep_prob: 1.0,
                },
            )

        dict_result = {}
        for no in range(len(self._label)):
            dict_result[self._label[no]] = []
            for k in range(probs.shape[0]):
                dict_result[self._label[no]].append(probs[k, no])
        return dict_result



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

from pathlib import Path
import os

home = str(Path.home()) + '/Malaya'
version = '0.9'
bump_version = '0.9.0.0'
version_path = home + '/version'


def delete_folder(home):
    for root, dirs, files in os.walk(home):
        for file in files:
            os.remove(os.path.join(root, file))


try:
    if not os.path.exists(home):
        os.makedirs(home)
except:
    print('cannot make directory for caching, exiting.')
    sys.exit(1)

if not os.path.isfile(version_path):
    print('not found any version, deleting previous version models..')
    delete_folder(home)
    with open(version_path, 'w') as fopen:
        fopen.write(version)
else:
    with open(version_path, 'r') as fopen:
        cached_version = fopen.read()
    if version not in cached_version:
        print('deleting previous version models..')
        delete_folder(home)
        with open(version_path, 'w') as fopen:
            fopen.write(version)


def delete_cache():
    """
    Remove cached data, this is will delete entire cache folder. Selected items to delete will be implement soon.
    """
    try:
        print('deleting cached models..')
        delete_folder(home)
        with open(version_path, 'w') as fopen:
            fopen.write(version)
        print('Done.')
    except:
        print(
            'failed to clear cached models. Please make sure %s is able to overwrite from Malaya'
            % (home)
        )


from .entities import crf_entities, deep_entities, get_available_entities_models
from .language_detection import (
    multinomial_detect_languages,
    xgb_detect_languages,
    get_language_labels,
    sgd_detect_languages,
)
from .normalizer import (
    spell_normalizer,
    fuzzy_normalizer,
    load_malay_dictionary,
    basic_normalizer,
    deep_normalizer,
)
from .num2word import (
    to_cardinal,
    to_ordinal,
    to_ordinal_num,
    to_currency,
    to_year,
)
from .pos_entities import deep_pos_entities, get_available_pos_entities_models
from .pos import naive_pos, crf_pos, deep_pos, get_available_pos_models
from .sentiment import (
    deep_sentiment,
    bayes_sentiment,
    pretrained_bayes_sentiment,
    get_available_sentiment_models,
    pretrained_xgb_sentiment,
)
from .spell import naive_speller
from .stemmer import naive_stemmer, sastrawi_stemmer, deep_stemmer
from .summarization import (
    summarize_lsa,
    summarize_nmf,
    summarize_lda,
    summarize_deep_learning,
)
from .text_functions import voting_stack
from .topic_modelling import (
    lda_topic_modelling,
    nmf_topic_modelling,
    lsa_topic_modelling,
    lda2vec_topic_modelling,
)
from .topics_influencers import (
    load_internal_data,
    fuzzy_get_influencers,
    fuzzy_get_topics,
    fuzzy_get_location,
    fast_get_topics,
    fast_get_influencers,
    deep_get_topics,
    deep_get_influencers,
    deep_siamese_get_topics,
    deep_siamese_get_influencers,
)
from .toxic import multinomial_detect_toxic, logistics_detect_toxic, deep_toxic
from .word2vec import malaya_word2vec, Word2Vec


def describe_pos_malaya():
    """
    Describe Malaya Part-Of-Speech supported (deprecated, use describe_pos() instead).
    """
    print(
        'This classes are deprecated, we prefer to use `malaya.describe_pos()`'
    )
    print('KT - Kata Tanya')
    print('KJ - Kata Kerja')
    print('KP - Kata Perintah')
    print('KPA - Kata Pangkal')
    print('KB - Kata Bantu')
    print('KPENGUAT - Kata Penguat')
    print('KPENEGAS - Kata Penegas')
    print('NAFI - Kata Nafi')
    print('KPEMERI - Kata Pemeri')
    print('KS - Kata Sendi')
    print('KPEMBENAR - Kata Pembenar')
    print('NAFI - Kata Nafi')
    print('NO - Numbers')
    print('SUKU - Suku Bilangan')
    print('PISAHAN - Kata Pisahan')
    print('KETERANGAN - Kata Keterangan')
    print('ARAH - Kata Arah')
    print('KH - Kata Hubung')
    print('GN - Ganti Nama')
    print('KA - Kata Adjektif')
    print('O - not related, out scope')


def describe_pos():
    """
    Describe Part-Of-Speech supported.
    """
    print('ADJ - Adjective, kata sifat')
    print('ADP - Adposition')
    print('ADV - Adverb, kata keterangan')
    print('ADX - Auxiliary verb, kata kerja tambahan')
    print('CCONJ - Coordinating conjuction, kata hubung')
    print('DET - Determiner, kata penentu')
    print('NOUN - Noun, kata nama')
    print('NUM - Number, nombor')
    print('PART - Particle')
    print('PRON - Pronoun, kata ganti')
    print('PROPN - Proper noun, kata ganti nama khas')
    print('SCONJ - Subordinating conjunction')
    print('SYM - Symbol')
    print('VERB - Verb, kata kerja')
    print('X - Other')


def describe_entities_malaya():
    """
    Describe Malaya Entities supported (deprecated, use describe_entities() instead).
    """
    print(
        'This classes are deprecated, we prefer to use `malaya.describe_entities()`'
    )
    print('PRN - person, group of people, believes, etc')
    print('LOC - location')
    print('NORP - Military, police, government, Parties, etc')
    print('ORG - Organization, company')
    print('LAW - related law document, etc')
    print('ART - art of work, special names, etc')
    print('EVENT - event happen, etc')
    print('FAC - facility, hospitals, clinics, etc')
    print('TIME - date, day, time, etc')
    print('O - not related, out scope')


def describe_entities():
    """
    Describe Entities supported.
    """
    print('OTHER - Other')
    print('law - law, regulation, related law documents, documents, etc')
    print('location - location, place')
    print('organization - organization, company, government, facilities, etc')
    print('person - person, group of people, believes, etc')
    print('quantity - numbers, quantity')
    print('time - date, day, time, etc')
    print('event - unique event happened, etc')



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import re
import tensorflow as tf
import numpy as np
import os
import json
from . import home
from .text_functions import (
    process_word_pos_entities,
    char_str_idx,
    generate_char_seq,
)
from .utils import load_graph, download_file

char_settings = home + '/char-settings.json'
char_frozen = home + '/char_frozen_model.pb'
concat_settings = home + '/concat-settings.json'
concat_frozen = home + '/concat_frozen_model.pb'
attention_settings = home + '/attention-settings.json'
attention_frozen = home + '/attention_pos_frozen_model.pb'


class DEEP_MODELS:
    def __init__(self, nodes, sess, predict):
        self.nodes = nodes
        self.sess = sess
        self.__predict = predict

    def predict(self, string):
        assert isinstance(string, str), 'input must be a string'
        return self.__predict(string, self.sess, self.nodes)


def get_entity_char(string, sess, model):
    batch_x = char_str_idx(
        [process_word_pos_entities(w) for w in string.split()],
        model['char2idx'],
        0,
    )
    logits, logits_pos = sess.run(
        [tf.argmax(model['logits'], 1), tf.argmax(model['logits_pos'], 1)],
        feed_dict = {model['X']: batch_x},
    )
    results = []
    for no, i in enumerate(string.split()):
        results.append(
            (
                i,
                model['idx2tag'][str(logits[no])],
                model['idx2pos'][str(logits_pos[no])],
            )
        )
    return results


def get_entity_concat(string, sess, model):
    array_X = char_str_idx(
        [[process_word_pos_entities(w) for w in string.split()]],
        model['word2idx'],
        2,
    )
    batch_x_char = generate_char_seq(
        array_X, model['idx2word'], model['char2idx']
    )
    Y_pred, Y_pos = sess.run(
        [model['crf_decode'], model['crf_decode_pos']],
        feed_dict = {
            model['word_ids']: array_X,
            model['char_ids']: batch_x_char,
        },
    )
    results = []
    for no, i in enumerate(string.split()):
        results.append(
            (
                i,
                model['idx2tag'][str(Y_pred[0, no])],
                model['idx2pos'][str(Y_pos[0, no])],
            )
        )
    return results


def get_available_pos_entities_models():
    return ['char', 'concat', 'attention']


def deep_pos_entities(model = 'attention'):
    assert isinstance(model, str), 'model must be a string'
    if model == 'char':
        if not os.path.isfile(char_settings):
            print('downloading POS-ENTITIES char settings')
            download_file('char-settings-pos-entities.json', char_settings)
        with open(char_settings, 'r') as fopen:
            nodes = json.loads(fopen.read())
        if not os.path.isfile(char_frozen):
            print('downloading POS-ENTITIES frozen char model')
            download_file('char-frozen-model-pos-entities.pb', char_frozen)
        g = load_graph(char_frozen)
        nodes['X'] = g.get_tensor_by_name('import/Placeholder:0')
        nodes['logits'] = g.get_tensor_by_name('import/logits:0')
        nodes['logits_pos'] = g.get_tensor_by_name('import/logits_pos:0')
        return DEEP_MODELS(
            nodes, tf.InteractiveSession(graph = g), get_entity_char
        )
    elif model == 'concat':
        if not os.path.isfile(concat_settings):
            print('downloading POS-ENTITIES concat settings')
            download_file('concat-settings-pos-entities.json', concat_settings)
        with open(concat_settings, 'r') as fopen:
            nodes = json.loads(fopen.read())
        if not os.path.isfile(concat_frozen):
            print('downloading POS-ENTITIES frozen concat model')
            download_file('concat-frozen-model-pos-entities.pb', concat_frozen)
        g = load_graph(concat_frozen)
        nodes['word_ids'] = g.get_tensor_by_name('import/Placeholder:0')
        nodes['char_ids'] = g.get_tensor_by_name('import/Placeholder_1:0')
        nodes['crf_decode'] = g.get_tensor_by_name(
            'import/entity-logits/cond/Merge:0'
        )
        nodes['crf_decode_pos'] = g.get_tensor_by_name(
            'import/pos-logits/cond/Merge:0'
        )
        nodes['idx2word'] = {int(k): v for k, v in nodes['idx2word'].items()}
        return DEEP_MODELS(
            nodes, tf.InteractiveSession(graph = g), get_entity_concat
        )
    elif model == 'attention':
        if not os.path.isfile(attention_settings):
            print('downloading POS-ENTITIES attention settings')
            download_file(
                'attention-settings-pos-entities.json', attention_settings
            )
        with open(attention_settings, 'r') as fopen:
            nodes = json.loads(fopen.read())
        if not os.path.isfile(attention_frozen):
            print('downloading POS-ENTITIES frozen attention model')
            download_file(
                'attention-frozen-model-pos-entities.pb', attention_frozen
            )
        g = load_graph(attention_frozen)
        nodes['word_ids'] = g.get_tensor_by_name('import/Placeholder:0')
        nodes['char_ids'] = g.get_tensor_by_name('import/Placeholder_1:0')
        nodes['crf_decode'] = g.get_tensor_by_name(
            'import/entity-logits/cond/Merge:0'
        )
        nodes['crf_decode_pos'] = g.get_tensor_by_name(
            'import/pos-logits/cond/Merge:0'
        )
        nodes['idx2word'] = {int(k): v for k, v in nodes['idx2word'].items()}
        return DEEP_MODELS(
            nodes, tf.InteractiveSession(graph = g), get_entity_concat
        )
    else:
        raise Exception('model not supported')



from tqdm import tqdm
import tensorflow as tf
import requests


def download_file(url, filename):
    r = requests.get(
        'http://s3-ap-southeast-1.amazonaws.com/huseinhouse-storage/' + url,
        stream = True,
    )
    total_size = int(r.headers['content-length'])
    with open(filename, 'wb') as f:
        for data in tqdm(
            iterable = r.iter_content(chunk_size = 1048576),
            total = total_size / 1048576,
            unit = 'MB',
            unit_scale = True,
        ):
            f.write(data)


def load_graph(frozen_graph_filename):
    with tf.gfile.GFile(frozen_graph_filename, 'rb') as f:
        graph_def = tf.GraphDef()
        graph_def.ParseFromString(f.read())
    with tf.Graph().as_default() as graph:
        tf.import_graph_def(graph_def)
    return graph



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import json
import pickle
import os
import tensorflow as tf
from .utils import download_file, load_graph
from . import home
from .paths import PATH_TOXIC, S3_PATH_TOXIC
from .sklearn_model import TOXIC
from .tensorflow_model import DEEP_TOXIC


def get_available_toxic_models():
    """
    List available deep learning toxicity analysis models.
    """
    return ['bahdanau', 'hierarchical', 'luong', 'fast-text', 'entity-network']


def multinomial_detect_toxic():
    """
    Load multinomial toxic model.

    Returns
    -------
    TOXIC : malaya.sklearn_model.TOXIC class
    """
    if not os.path.isfile(PATH_TOXIC['multinomial']['model']):
        print('downloading TOXIC pickled multinomial model')
        download_file(
            S3_PATH_TOXIC['multinomial']['model'],
            PATH_TOXIC['multinomial']['model'],
        )
    if not os.path.isfile(PATH_TOXIC['multinomial']['vector']):
        print('downloading TOXIC pickled tfidf vectorizations')
        download_file(
            S3_PATH_TOXIC['multinomial']['vector'],
            PATH_TOXIC['multinomial']['vector'],
        )
    with open(PATH_TOXIC['multinomial']['model'], 'rb') as fopen:
        multinomial = pickle.load(fopen)
    with open(PATH_TOXIC['multinomial']['vector'], 'rb') as fopen:
        vectorize = pickle.load(fopen)
    return TOXIC(multinomial, vectorize)


def logistics_detect_toxic():
    """
    Load logistic toxic model.

    Returns
    -------
    TOXIC : malaya.sklearn_model.TOXIC class
    """
    if not os.path.isfile(PATH_TOXIC['logistic']['model']):
        print('downloading TOXIC pickled logistics model')
        download_file(
            S3_PATH_TOXIC['logistic']['model'], PATH_TOXIC['logistic']['model']
        )
    if not os.path.isfile(PATH_TOXIC['logistic']['vector']):
        print('downloading TOXIC pickled tfidf vectorizations')
        download_file(
            S3_PATH_TOXIC['logistic']['vector'],
            PATH_TOXIC['logistic']['vector'],
        )
    with open(PATH_TOXIC['logistic']['model'], 'rb') as fopen:
        logistic = pickle.load(fopen)
    with open(PATH_TOXIC['logistic']['vector'], 'rb') as fopen:
        vectorize = pickle.load(fopen)
    return TOXIC(logistic, vectorize)


def deep_toxic(model = 'luong'):
    """
    Load deep learning sentiment analysis model.

    Parameters
    ----------
    model : str, optional (default='luong')
        Model architecture supported. Allowed values:

        * ``'fast-text'`` - Fast-text architecture, embedded and logits layers only
        * ``'hierarchical'`` - LSTM with hierarchical attention architecture
        * ``'bahdanau'`` - LSTM with bahdanau attention architecture
        * ``'luong'`` - LSTM with luong attention architecture
        * ``'entity-network'`` - Recurrent Entity-Network architecture

    Returns
    -------
    TOXIC: malaya.tensorflow_model.DEEP_TOXIC class
    """
    assert isinstance(model, str), 'model must be a string'
    model = model.lower()
    if model == 'fast-text':
        if not os.path.isfile(PATH_TOXIC['fast-text']['model']):
            print('downloading TOXIC frozen fast-text model')
            download_file(
                S3_PATH_TOXIC['fast-text']['model'],
                PATH_TOXIC['fast-text']['model'],
            )
        if not os.path.isfile(PATH_TOXIC['fast-text']['setting']):
            print('downloading TOXIC fast-text dictionary')
            download_file(
                S3_PATH_TOXIC['fast-text']['setting'],
                PATH_TOXIC['fast-text']['setting'],
            )
        if not os.path.isfile(PATH_TOXIC['fast-text']['pickle']):
            print('downloading TOXIC fast-text bigrams')
            download_file(
                S3_PATH_TOXIC['fast-text']['pickle'],
                PATH_TOXIC['fast-text']['pickle'],
            )
        with open(PATH_TOXIC['fast-text']['setting'], 'r') as fopen:
            dictionary = json.load(fopen)['dictionary']
        with open(PATH_TOXIC['fast-text']['pickle'], 'rb') as fopen:
            ngram = pickle.load(fopen)
        g = load_graph(PATH_TOXIC['fast-text']['model'])
        return DEEP_TOXIC(
            g.get_tensor_by_name('import/Placeholder:0'),
            g.get_tensor_by_name('import/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            ngram = ngram,
        )
    elif model == 'hierarchical':
        if not os.path.isfile(PATH_TOXIC['hierarchical']['model']):
            print('downloading TOXIC frozen hierarchical model')
            download_file(
                S3_PATH_TOXIC['hierarchical']['model'],
                PATH_TOXIC['hierarchical']['model'],
            )
        if not os.path.isfile(PATH_TOXIC['hierarchical']['setting']):
            print('downloading TOXIC hierarchical dictionary')
            download_file(
                S3_PATH_TOXIC['hierarchical']['setting'],
                PATH_TOXIC['hierarchical']['setting'],
            )
        with open(PATH_TOXIC['hierarchical']['setting'], 'r') as fopen:
            dictionary = json.load(fopen)['dictionary']
        g = load_graph(PATH_TOXIC['hierarchical']['model'])
        return DEEP_TOXIC(
            g.get_tensor_by_name('import/Placeholder:0'),
            g.get_tensor_by_name('import/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            alphas = g.get_tensor_by_name('import/alphas:0'),
        )
    elif model == 'bahdanau':
        if not os.path.isfile(PATH_TOXIC['bahdanau']['model']):
            print('downloading TOXIC frozen bahdanau model')
            download_file(
                S3_PATH_TOXIC['bahdanau']['model'],
                PATH_TOXIC['bahdanau']['model'],
            )
        if not os.path.isfile(PATH_TOXIC['bahdanau']['setting']):
            print('downloading TOXIC bahdanau dictionary')
            download_file(
                S3_PATH_TOXIC['bahdanau']['setting'],
                PATH_TOXIC['bahdanau']['setting'],
            )
        with open(PATH_TOXIC['bahdanau']['setting'], 'r') as fopen:
            dictionary = json.load(fopen)['dictionary']
        g = load_graph(PATH_TOXIC['bahdanau']['model'])
        return DEEP_TOXIC(
            g.get_tensor_by_name('import/Placeholder:0'),
            g.get_tensor_by_name('import/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            alphas = g.get_tensor_by_name('import/alphas:0'),
        )
    elif model == 'luong':
        if not os.path.isfile(PATH_TOXIC['luong']['model']):
            print('downloading TOXIC frozen luong model')
            download_file(
                S3_PATH_TOXIC['luong']['model'], PATH_TOXIC['luong']['model']
            )
        if not os.path.isfile(PATH_TOXIC['luong']['setting']):
            print('downloading TOXIC luong dictionary')
            download_file(
                S3_PATH_TOXIC['luong']['setting'],
                PATH_TOXIC['luong']['setting'],
            )
        with open(PATH_TOXIC['luong']['setting'], 'r') as fopen:
            dictionary = json.load(fopen)
        g = load_graph(PATH_TOXIC['luong']['model'])
        return DEEP_TOXIC(
            g.get_tensor_by_name('import/Placeholder:0'),
            g.get_tensor_by_name('import/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            alphas = g.get_tensor_by_name('import/alphas:0'),
        )
    elif model == 'entity-network':
        if not os.path.isfile(PATH_TOXIC['entity-network']['model']):
            print('downloading TOXIC frozen entity-network model')
            download_file(
                S3_PATH_TOXIC['entity-network']['model'],
                PATH_TOXIC['entity-network']['model'],
            )
        if not os.path.isfile(PATH_TOXIC['entity-network']['setting']):
            print('downloading TOXIC entity-network dictionary')
            download_file(
                S3_PATH_TOXIC['entity-network']['setting'],
                PATH_TOXIC['entity-network']['setting'],
            )
        with open(PATH_TOXIC['entity-network']['setting'], 'r') as fopen:
            dictionary = json.load(fopen)
        g = load_graph(PATH_TOXIC['entity-network']['model'])
        return DEEP_TOXIC(
            g.get_tensor_by_name('import/Placeholder_question:0'),
            g.get_tensor_by_name('import/logits:0'),
            tf.InteractiveSession(graph = g),
            model,
            dictionary,
            dropout_keep_prob = g.get_tensor_by_name(
                'import/Placeholder_dropout_keep_prob:0'
            ),
            story = g.get_tensor_by_name('import/Placeholder_story:0'),
        )
    else:
        raise Exception(
            'model sentiment not supported, please check supported models from malaya.get_available_toxic_models()'
        )



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import pickle
import os
import json
import tensorflow as tf
from .utils import download_file, load_graph
from .text_functions import entities_textcleaning
from .paths import PATH_ENTITIES, S3_PATH_ENTITIES
from .tensorflow_model import TAGGING
from .sklearn_model import CRF


def get_available_entities_models():
    """
    List available deep learning entities models, ['concat', 'bahdanau', 'luong', 'entity-network', 'attention']
    """
    return ['concat', 'bahdanau', 'luong', 'entity-network', 'attention']


def crf_entities():
    """
    Load CRF Entities Recognition model.

    Returns
    -------
    CRF : malaya.sklearn_model.CRF class
    """
    if not os.path.isfile(PATH_ENTITIES['crf']['model']):
        print('downloading ENTITIES frozen CRF model')
        download_file(
            S3_PATH_ENTITIES['crf']['model'], PATH_ENTITIES['crf']['model']
        )
    with open(PATH_ENTITIES['crf']['model'], 'rb') as fopen:
        model = pickle.load(fopen)
    return CRF(model)


def deep_entities(model = 'bahdanau'):
    """
    Load deep learning NER model.

    Parameters
    ----------
    model : str, optional (default='bahdanau')
        Model architecture supported. Allowed values:

        * ``'concat'`` - Concating character and word embedded for BiLSTM
        * ``'bahdanau'`` - Concating character and word embedded including Bahdanau Attention for BiLSTM
        * ``'luong'`` - Concating character and word embedded including Luong Attention for BiLSTM
        * ``'entity-network'`` - Concating character and word embedded on hybrid Entity-Network and RNN

    Returns
    -------
    TAGGING: malaya.tensorflow_model.TAGGING class
    """
    assert isinstance(model, str), 'model must be a string'
    model = model.lower()
    if model in ['concat', 'bahdanau', 'luong', 'entity-network']:
        if not os.path.isfile(PATH_ENTITIES[model]['model']):
            print('downloading ENTITIES frozen %s model' % (model))
            download_file(
                S3_PATH_ENTITIES[model]['model'], PATH_ENTITIES[model]['model']
            )
        if not os.path.isfile(PATH_ENTITIES[model]['setting']):
            print('downloading ENTITIES %s dictionary' % (model))
            download_file(
                S3_PATH_ENTITIES[model]['setting'],
                PATH_ENTITIES[model]['setting'],
            )
        with open(PATH_ENTITIES[model]['setting'], 'r') as fopen:
            nodes = json.loads(fopen.read())
        g = load_graph(PATH_ENTITIES[model]['model'])
        if model == 'entity-network':
            return TAGGING(
                g.get_tensor_by_name('import/question:0'),
                g.get_tensor_by_name('import/char_ids:0'),
                g.get_tensor_by_name('import/logits:0'),
                nodes,
                tf.InteractiveSession(graph = g),
                model,
                story = g.get_tensor_by_name('import/story:0'),
            )
        else:
            return TAGGING(
                g.get_tensor_by_name('import/Placeholder:0'),
                g.get_tensor_by_name('import/Placeholder_1:0'),
                g.get_tensor_by_name('import/logits:0'),
                nodes,
                tf.InteractiveSession(graph = g),
                model,
            )

    else:
        raise Exception(
            'model not supported, please check supported models from malaya.get_available_entities_models()'
        )
